# coding=utf-8
# Copyright 2019 The TensorFlow Datasets Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""CBIS-DDSM mammography dataset."""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import csv
import os
import re

from absl import logging
import numpy as np
import tensorflow as tf
import tensorflow_datasets.public_api as tfds


_DESCRIPTION = """\
The CBIS-DDSM (Curated Breast Imaging Subset of DDSM) is an updated and
standardized version of the Digital Database for Screening Mammography (DDSM).
The DDSM is a database of 2,620 scanned film mammography studies.
It contains normal, benign, and malignant cases with verified pathology
information.

The default config is made of patches extracted from the original mammograms,
following the description from http://arxiv.org/abs/1708.09427, in order to
frame the task to solve in a traditional image classification setting.

Because special software and libraries are needed to download and read the
images contained in the dataset, TFDS assumes that the user has downloaded the
original DCIM files and converted them to PNG.

The following commands (or equivalent) should be used to generate the PNG files,
in order to guarantee reproducible results:

  find $DATASET_DCIM_DIR -name '*.dcm' | \\
  xargs -n1 -P8 -I{} bash -c 'f={}; dcmj2pnm $f | convert - ${f/.dcm/.png}'
"""

_CITATION = """\
@misc{CBIS_DDSM_Citation,
  doi = {10.7937/k9/tcia.2016.7o02s9cy},
  url = {https://wiki.cancerimagingarchive.net/x/lZNXAQ},
  author = {Sawyer-Lee,  Rebecca and Gimenez,  Francisco and Hoogi,  Assaf and Rubin,  Daniel},
  title = {Curated Breast Imaging Subset of DDSM},
  publisher = {The Cancer Imaging Archive},
  year = {2016},
}
@article{TCIA_Citation,
  author = {
    K. Clark and B. Vendt and K. Smith and J. Freymann and J. Kirby and
    P. Koppel and S. Moore and S. Phillips and D. Maffitt and M. Pringle and
    L. Tarbox and F. Prior
  },
  title = {{The Cancer Imaging Archive (TCIA): Maintaining and Operating a
  Public Information Repository}},
  journal = {Journal of Digital Imaging},
  volume = {26},
  month = {December},
  year = {2013},
  pages = {1045-1057},
}
@article{DBLP:journals/corr/abs-1708-09427,
  author    = {Li Shen},
  title     = {End-to-end Training for Whole Image Breast Cancer Diagnosis using
               An All Convolutional Design},
  journal   = {CoRR},
  volume    = {abs/1708.09427},
  year      = {2017},
  url       = {http://arxiv.org/abs/1708.09427},
  archivePrefix = {arXiv},
  eprint    = {1708.09427},
  timestamp = {Mon, 13 Aug 2018 16:48:35 +0200},
  biburl    = {https://dblp.org/rec/bib/journals/corr/abs-1708-09427},
  bibsource = {dblp computer science bibliography, https://dblp.org}
}
"""

_CALC_TEST_CSV_URL = 'https://wiki.cancerimagingarchive.net/download/attachments/22516629/calc_case_description_test_set.csv?version=1&modificationDate=1506796343686&api=v2'
_CALC_TRAIN_CSV_URL = 'https://wiki.cancerimagingarchive.net/download/attachments/22516629/calc_case_description_train_set.csv?version=1&modificationDate=1506796349666&api=v2'
_MASS_TEST_CSV_URL = 'https://wiki.cancerimagingarchive.net/download/attachments/22516629/mass_case_description_test_set.csv?version=1&modificationDate=1506796343175&api=v2'
_MASS_TRAIN_CSV_URL = 'https://wiki.cancerimagingarchive.net/download/attachments/22516629/mass_case_description_train_set.csv?version=1&modificationDate=1506796355038&api=v2'

_IMAGE_VIEW_LABELS = (
    'CC',
    'MLO',
)
_BREAST_LABELS = (
    'LEFT',
    'RIGHT',
)
_BREAST_DENSITY_NUM_CLASSES = 4  # Original range: [1, 4]
_PATHOLOGY_LABELS = (
    'BENIGN',
    'BENIGN_WITHOUT_CALLBACK',
    'MALIGNANT',
)
_ASSESSMENT_NUM_CLASSES = 6  # Original range: [0, 5]
_SUBTELTY_NUM_CLASSES = 6  # Original range: [0, 5]
_DCIM_REGEX = re.compile(
    r'^.*/(?P<study>1.3.6.1.4.1.9590.100.1.2.[0-9.]+)/(?P<series>1.3.6.1.4.1.9590.100.1.2.[0-9.]+)/(?P<instance>.+).dcm$'
)


class CuratedBreastImagingDDSMConfig(tfds.core.BuilderConfig):
  """BuilderConfig for CuratedBreastImagingDDSM."""

  def __init__(self, image_size=None, patch_size=None, **kwargs):
    kwargs['supported_versions'] = [
        tfds.core.Version('2.0.0'),
        tfds.core.Version('1.0.0'),
    ]
    super(CuratedBreastImagingDDSMConfig, self).__init__(**kwargs)
    self.image_size = image_size
    self.patch_size = patch_size


class CuratedBreastImagingDDSM(tfds.core.GeneratorBasedBuilder):
  """Curated Breast Imaging Subset of DDSM."""

  BUILDER_CONFIGS = [
      CuratedBreastImagingDDSMConfig(
          name='patches',
          version=tfds.core.Version(
              '0.2.0', experiments={tfds.core.Experiment.S3: False}),
          description=('Patches containing both calsification and mass cases, '
                       'plus pathces with no abnormalities. Designed as a '
                       'traditional 5-class classification task.'),
          image_size=(1152, 896),  # Note: (height, width).
          patch_size=(224, 224)),
      CuratedBreastImagingDDSMConfig(
          name='original-calc',
          version=tfds.core.Version(
              '0.1.0', experiments={tfds.core.Experiment.S3: False}),
          description=('Original images of the calcification cases compressed '
                       'in lossless PNG.')),
      CuratedBreastImagingDDSMConfig(
          name='original-mass',
          version=tfds.core.Version(
              '0.1.0', experiments={tfds.core.Experiment.S3: False}),
          description=('Original images of the mass cases compressed in '
                       'lossless PNG.')),
  ]

  def _info(self):
    features_fn_map = {
        'original-calc': self._get_features_original_calc,
        'original-mass': self._get_features_original_mass,
        'patches': self._get_features_patches
    }
    if self.builder_config.name not in features_fn_map:
      raise ValueError('Builder config named {} not supported!'.format(
          self.builder_config.name))

    return tfds.core.DatasetInfo(
        builder=self,
        description=_DESCRIPTION,
        features=features_fn_map[self.builder_config.name](),
        urls=['https://wiki.cancerimagingarchive.net/display/Public/CBIS-DDSM'],
        citation=_CITATION)

  def _get_features_original_base(self):
    return {
        'id': tfds.features.Text(),
        'breast': tfds.features.ClassLabel(names=_BREAST_LABELS),
        'image': tfds.features.Image(shape=(None, None, 1)),
        'view': tfds.features.ClassLabel(names=_IMAGE_VIEW_LABELS),
        'patient': tfds.features.Text(),
        'abnormalities': {
            'id': tfds.features.Tensor(shape=(), dtype=tf.int32),
            'mask': tfds.features.Image(shape=(None, None, 1)),
            'assessment':
                tfds.features.ClassLabel(num_classes=_ASSESSMENT_NUM_CLASSES),
            'pathology': tfds.features.ClassLabel(names=_PATHOLOGY_LABELS),
            'subtlety':
                tfds.features.ClassLabel(num_classes=_SUBTELTY_NUM_CLASSES),
            # TODO(jpuigcerver): Include original crops when TFDS allows it.
            # The problem seems to be in the compute statistics steps, since
            # a given example may have crops of different sizes and this is
            # not handled properly.
            # 'crop': tfds.features.Image(shape=(None, None, 1)),
        }
    }

  def _get_features_original_calc(self):
    features = self._get_features_original_base()
    features['abnormalities'].update({
        'calc_type':
            tfds.features.ClassLabel(
                names_file=tfds.core.get_tfds_path(
                    os.path.join('image', 'cbis_ddsm_calc_types.txt'))),
        'calc_distribution':
            tfds.features.ClassLabel(
                names_file=tfds.core.get_tfds_path(
                    os.path.join('image', 'cbis_ddsm_calc_distributions.txt'))),
    })
    features['abnormalities'] = tfds.features.Sequence(
        tfds.features.FeaturesDict(features['abnormalities']))
    return tfds.features.FeaturesDict(features)

  def _get_features_original_mass(self):
    features = self._get_features_original_base()
    features['abnormalities'].update({
        'mass_shape':
            tfds.features.ClassLabel(
                names_file=tfds.core.get_tfds_path(
                    os.path.join('image', 'cbis_ddsm_mass_shapes.txt'))),
        'mass_margins':
            tfds.features.ClassLabel(
                names_file=tfds.core.get_tfds_path(
                    os.path.join('image', 'cbis_ddsm_mass_margins.txt'))),
    })
    features['abnormalities'] = tfds.features.Sequence(
        tfds.features.FeaturesDict(features['abnormalities']))
    return tfds.features.FeaturesDict(features)

  def _get_features_patches(self):
    return tfds.features.FeaturesDict({
        'id': tfds.features.Text(),
        'image':
            tfds.features.Image(shape=(None, None, 1), encoding_format='jpeg'),
        'label':
            tfds.features.ClassLabel(
                names_file=tfds.core.get_tfds_path(
                    os.path.join('image', 'cbis_ddsm_patch_labels.txt'))),
    })

  def _split_generators(self, dl_manager):
    if self.builder_config.name in ['original-calc', 'original-mass']:
      return self._split_generators_original(dl_manager)
    elif self.builder_config.name == 'patches':
      return self._split_generators_patches(dl_manager)
    else:
      raise ValueError('Builder config named {} not supported!'.format(
          self.builder_config.name))

  def _split_generators_original(self, dl_manager):
    if self.builder_config.name == 'original-calc':
      test_url = _CALC_TEST_CSV_URL
      train_url = _CALC_TRAIN_CSV_URL
    elif self.builder_config.name == 'original-mass':
      test_url = _MASS_TEST_CSV_URL
      train_url = _MASS_TRAIN_CSV_URL
    else:
      raise ValueError('Builder config named {} not supported!'.format(
          self.builder_config.name))

    resources = {'test': test_url, 'train': train_url}
    resource_paths = dl_manager.download_and_extract(resources)
    patients_data = _load_csv_files(dl_manager.manual_dir, resource_paths)

    return [
        tfds.core.SplitGenerator(
            name=tfds.Split.TRAIN,
            num_shards=10,
            gen_kwargs={
                'generate_fn': self._generate_examples_original,
                'patients_data': patients_data,
                'yield_from_train_csv': True,  # Yield train examples.
            },
        ),
        tfds.core.SplitGenerator(
            name=tfds.Split.TEST,
            num_shards=1,
            gen_kwargs={
                'generate_fn': self._generate_examples_original,
                'patients_data': patients_data,
                'yield_from_train_csv': False,  # Yield test examples.
            },
        ),
    ]

  def _split_generators_patches(self, dl_manager):
    resources_urls = {
        'calc-test': _CALC_TEST_CSV_URL,
        'calc-train': _CALC_TRAIN_CSV_URL,
        'mass-test': _MASS_TEST_CSV_URL,
        'mass-train': _MASS_TRAIN_CSV_URL
    }
    resources = {
        key: tfds.download.Resource(
            url=url, extract_method=tfds.download.ExtractMethod.NO_EXTRACT)
        for key, url in resources_urls.items()
    }
    resource_paths = dl_manager.download_and_extract(resources)
    patients_data = _load_csv_files(dl_manager.manual_dir, resource_paths)

    # Statistics about the resulting splits.
    # Whole dataset:
    #   Num patients: 1566, of which have malignant abnormalities: 48.0%
    #   Num mamographies: 3103, of which have malignant abnormalities: 44.3%
    #   Num abnormalities: 3568, of which are malignant: 40.8%
    # Test split:
    #   Num patients: 234, of which have malignant abnormalities: 47.0%
    #   Num mamographies: 450, of which have malignant abnormalities: 44.2%
    #   Num abnormalities: 538, of which are malignant: 39.8%
    # Train split:
    #   Num patients: 1197, of which have malignant abnormalities: 48.0%
    #   Num mamographies: 2386, of which have malignant abnormalities: 44.1%
    #   Num abnormalities: 2722, of which are malignant: 41.0%
    # Validation split:
    #   Num patients: 135, of which have malignant abnormalities: 49.6%
    #   Num mamographies: 267, of which have malignant abnormalities: 46.4%
    #   Num abnormalities: 308, of which are malignant: 41.2%
    patients_test, patients_train, patients_valid = _split_patients(
        patients_data)
    patients_data_test = _select_patients_data(patients_data, patients_test)
    patients_data_train = _select_patients_data(patients_data, patients_train)
    patients_data_valid = _select_patients_data(patients_data, patients_valid)

    return [
        tfds.core.SplitGenerator(
            name=tfds.Split.TRAIN,
            num_shards=1,
            gen_kwargs={
                'generate_fn': self._generate_examples_patches,
                'patients_data': patients_data_train,
                'image_size': self.builder_config.image_size,
                'patch_size': self.builder_config.patch_size,
            },
        ),
        tfds.core.SplitGenerator(
            name=tfds.Split.TEST,
            num_shards=1,
            gen_kwargs={
                'generate_fn': self._generate_examples_patches,
                'patients_data': patients_data_test,
                'image_size': self.builder_config.image_size,
                'patch_size': self.builder_config.patch_size,
            },
        ),
        tfds.core.SplitGenerator(
            name=tfds.Split.VALIDATION,
            num_shards=1,
            gen_kwargs={
                'generate_fn': self._generate_examples_patches,
                'patients_data': patients_data_valid,
                'image_size': self.builder_config.image_size,
                'patch_size': self.builder_config.patch_size,
            },
        ),
    ]

  def _generate_examples(self, generate_fn, **kwargs):
    """Yields examples."""
    return generate_fn(**kwargs)

  def _generate_examples_original(self, patients_data, yield_from_train_csv):

    def _include_example_in_split(example):
      if yield_from_train_csv:
        return example['csv_key'] == 'train'
      else:
        return example['csv_key'] == 'test'

    for _, patient_examples in sorted(patients_data.items()):
      for _, example in sorted(patient_examples.items()):
        if _include_example_in_split(example):
          record = {
              'id': example['id'],
              'patient': example['patient'],
              'image': example['image'],
              'view': example['view'],
              'breast': example['breast'],
              # pylint: disable=g-complex-comprehension
              'abnormalities': [{
                  k: v for k, v in abnormality.items() if k not in ['type']
              } for abnormality in example['abnormalities']],
              # pylint: enable=g-complex-comprehension
          }
          yield example['id'], record

  def _generate_examples_patches(self,
                                 patients_data,
                                 image_size=(1152, 896),
                                 patch_size=(224, 224),
                                 num_positive_patches_per_abnormality=10,
                                 num_background_patches_per_image=10):
    # Set random seed so that we always get the same patches in each split.
    np.random.seed(0x12345 + len(patients_data))

    for _, patient_examples in sorted(patients_data.items()):
      for _, example in sorted(patient_examples.items()):
        # Read the mammography image.
        image = _read_image(example['image'], image_size=image_size)
        abnormalities_masks = [
            _read_image(abnormality['mask'], image_size=image.shape)
            for abnormality in example['abnormalities']
        ]
        abnormalities_areas = [np.sum(mask > 0) for mask in abnormalities_masks]
        # Sample positive (abnormal) patches from the given mammography.
        for abnormality, abnormality_mask, abnormality_area in zip(
            example['abnormalities'], abnormalities_masks, abnormalities_areas):
          # Determine label for the given abnormality.
          if abnormality['pathology'].startswith('MALIGNANT'):
            benign_or_malignant = 'MALIGNANT'
          else:
            benign_or_malignant = 'BENING'
          if abnormality['type'] == 'calc':
            label = benign_or_malignant + '_CALCIFICATION'
          elif abnormality['type'] == 'mass':
            label = benign_or_malignant + '_MASS'
          else:
            raise ValueError('Unknown abnormality type: %r' %
                             abnormality['type'])
          # Sample positive patches from the given abnormality.
          for k, patch in enumerate(
              _sample_positive_patches(image, abnormality['mask'],
                                       abnormality_mask, abnormality_area,
                                       patch_size,
                                       num_positive_patches_per_abnormality)):
            patch_id = ('%s/abnorm_%s/patch_%d' %
                        (example['id'], abnormality['id'], k))
            record = {
                'id': patch_id,
                # Note: TFDS needs the shape to be (?, ?, 1).
                'image': np.expand_dims(patch, axis=-1),
                'label': label,
            }
            yield patch_id, record

        # Sample background patches from the given mammography.
        for k, patch in enumerate(
            _sample_negative_patches(image, example['image'],
                                     abnormalities_masks, abnormalities_areas,
                                     patch_size,
                                     num_background_patches_per_image)):
          id_ = '%s/background_%d' % (example['id'], k)
          record = {
              'id': id_,
              # Note: TFDS needs the shape to be (?, ?, 1).
              'image': np.expand_dims(patch, axis=-1),
              'label': 'BACKGROUND',
          }
          yield id_, record


def _load_csv_files(manual_dir, dictionary_of_csv_files):
  """Load the ground-truth data from the given dictionary of CSV files.

  Args:
    manual_dir: Path of the directory containing the images.
    dictionary_of_csv_files: Dictionary containing the key and filepath of each
      CSV file to load.

  Returns:
    A dictionary containing the ground-truth loaded from the CSV files.
  """
  # Data maps patients -> examples -> list of abnormalities
  data = {}
  for csv_key, csv_path in sorted(dictionary_of_csv_files.items()):
    with tf.io.gfile.GFile(csv_path, 'r') as f:
      csv_reader = csv.DictReader(f)
      for i, row in enumerate(csv_reader, 2):
        row = {k: v.strip() for k, v in row.items()}  # Strip all cells.
        # Construct example ID from the study and series IDs.
        example_id = _DCIM_REGEX.sub(r'\g<study>/\g<series>',
                                     row['image file path'])
        # Get path to the
        for key in [
            'image file path', 'ROI mask file path', 'cropped image file path'
        ]:
          row[key] = row[key].replace('.dcm', '.png')
          row[key] = os.path.join(manual_dir, *row[key].split('/'))
          if not tf.io.gfile.exists(row[key]):
            raise ValueError('Error processing line %d from csv file %s: '
                             'Image %r does not exist!' %
                             (i, csv_path, row[key]))

        mask_file_path = row['ROI mask file path']
        crop_file_path = row['cropped image file path']
        full_image = _read_image(row['image file path'])
        mask_image = _read_image(mask_file_path)
        crop_image = _read_image(crop_file_path)
        if full_image.shape == crop_image.shape:
          # TODO(jpuigcerver): THIS ASSUMES THAT THE CROP/MASK COLUMNS ARE JUST
          # REVERSED. I've checked that this is the case for a couple of rows,
          # but this issue happens a lot across all CSV files. Contact the
          # owners of the dataset to ask about this problem.
          mask_file_path, crop_file_path = crop_file_path, mask_file_path
        elif full_image.shape != mask_image.shape:
          # TODO(jpuigcerver): Contact the owners of the dataset to ask about
          # this problem.
          logging.error(
              'Error processing line %d from csv file %s: No suitable mask for '
              'the given image (expected size: %r, candidate sizes: %r). '
              'This abnormality will NOT be included in the dataset.', i,
              csv_path, full_image.shape, [mask_image.shape, crop_image.shape])
          continue

        abnormality = {
            'id': int(row['abnormality id']),
            'mask': mask_file_path,
            'assessment': row['assessment'],
            'pathology': row['pathology'],
            'subtlety': row['subtlety'],
        }
        if 'calc type' in row and 'calc distribution' in row:
          abnormality['type'] = 'calc'
          abnormality['calc_type'] = row['calc type']
          abnormality['calc_distribution'] = row['calc distribution']
        elif 'mass shape' in row and 'mass margins' in row:
          abnormality['type'] = 'mass'
          abnormality['mass_shape'] = row['mass shape']
          abnormality['mass_margins'] = row['mass margins']
        else:
          raise ValueError('CSV file is missing required columns.')

        example = {
            'id': example_id,
            'breast': row['left or right breast'],
            'patient': row['patient_id'],
            'image': row['image file path'],
            'view': row['image view'],
            'abnormalities': [abnormality],
            # Note: Useful to know whether the example is from train or test.
            'csv_key': csv_key,
        }
        _append_example_to_data(data, example)

  return data


def _append_example_to_data(data, example):
  """Append the given example to the data dictionary."""
  example_id = example['id']
  patient_id = example['patient']
  if patient_id in data:
    if example_id in data[example['patient']]:
      assert example_id == data[patient_id][example_id]['id']
      assert patient_id == data[patient_id][example_id]['patient']
      assert example['breast'] == data[patient_id][example_id]['breast']
      assert example['image'] == data[patient_id][example_id]['image']
      assert example['view'] == data[patient_id][example_id]['view']
      data[patient_id][example_id]['abnormalities'].extend(
          example['abnormalities'])
    else:
      data[example['patient']][example['id']] = example
  else:
    data[example['patient']] = {example['id']: example}


def _split_patients(data,
                    test_fraction=0.15,
                    train_fraction=0.765,
                    valid_fraction=0.085):
  """Split the patients in the data dictionary into test, train and valid sets."""
  assert test_fraction > 0 and train_fraction > 0 and valid_fraction > 0
  assert np.abs(test_fraction + train_fraction + valid_fraction - 1.0) < 1e-9
  all_patient_ids = sorted(list(data.keys()))
  np.random.seed(seed=0x12345)  # To make sure we always get the same splits.
  np.random.shuffle(all_patient_ids)
  cutoff_test = int(test_fraction * len(all_patient_ids))
  patients_test = all_patient_ids[:cutoff_test]
  cutoff_train = cutoff_test + int(train_fraction * len(all_patient_ids))
  patients_train = all_patient_ids[cutoff_test:cutoff_train]
  patients_valid = all_patient_ids[cutoff_train:]
  return set(patients_test), set(patients_train), set(patients_valid)


def _select_patients_data(data, patient_ids):
  return {k: v for k, v in data.items() if k in patient_ids}


def _read_image(filepath, image_size=None):
  """Read an image and optionally resize it (size must be: height, width)."""
  cv2 = tfds.core.lazy_imports.cv2
  with tf.io.gfile.GFile(filepath, 'rb') as f:
    image = cv2.imdecode(
        np.fromstring(f.read(), dtype=np.uint8), flags=cv2.IMREAD_GRAYSCALE)
    if image_size:
      # Note: cv2.resize actually expects (width, size).
      image = cv2.resize(image, (image_size[1], image_size[0]))
      assert image.shape == image_size
    return image


def _get_breast_mask(image, min_breast_color_threshold=0.05):
  """Get the binary mask of the breast region of the image."""
  cv2 = tfds.core.lazy_imports.cv2
  threshold = int(image.max() * min_breast_color_threshold)
  _, image_binary = cv2.threshold(image, threshold, 255, cv2.THRESH_BINARY)
  _, contours, _ = cv2.findContours(image_binary, cv2.RETR_LIST,
                                    cv2.CHAIN_APPROX_SIMPLE)
  contours_areas = [cv2.contourArea(cont) for cont in contours]
  biggest_contour_idx = np.argmax(contours_areas)
  return cv2.drawContours(
      np.zeros_like(image_binary), contours, biggest_contour_idx, 255,
      cv2.FILLED)


def _get_roi_from_mask(mask):
  cv2 = tfds.core.lazy_imports.cv2
  _, contours, _ = cv2.findContours(mask, cv2.RETR_LIST,
                                    cv2.CHAIN_APPROX_SIMPLE)
  contours_areas = [cv2.contourArea(cont) for cont in contours]
  biggest_contour_idx = np.argmax(contours_areas)
  return contours[biggest_contour_idx]


def _patch_overlaps_any_abnormality_above_threshold(y, x, patch_size,
                                                    abnormalities_masks,
                                                    abnormalities_areas,
                                                    min_overlap_threshold):
  """Return True if the given patch overlaps significantly with any abnormality.

  Given a patch and a single abnormality, the overlap between the two is
  significant if, and only if, the relative area of the intersection of the two
  w.r.t. the area of the patch is above `min_overlap_threshold` OR the
  area of the intersection w.r.t. the total abnormality area is above
  `min_overlap_threshold`.

  Args:
    y: Top-most coordinate of the patch.
    x: Left-most coordinate of the patch.
    patch_size: Tuple with (height, width) of the patch.
    abnormalities_masks: List with the binary mask of each abnormality.
    abnormalities_areas: List with the total area of each abnormality.
    min_overlap_threshold:

  Returns:
    Returns True if the above condition is met for any of the given
    abnormalities, or False otherwise.
  """
  patch_area = patch_size[0] * patch_size[1]
  for abnorm_mask, abnorm_area in zip(abnormalities_masks, abnormalities_areas):
    abnorm_in_patch_area = np.sum(
        abnorm_mask[y:(y + patch_size[0]), x:(x + patch_size[1])] > 0)
    abnorm_in_patch_wrt_patch = abnorm_in_patch_area / patch_area
    abnorm_in_patch_wrt_abnorm = abnorm_in_patch_area / abnorm_area
    if (abnorm_in_patch_wrt_patch > min_overlap_threshold or
        abnorm_in_patch_wrt_abnorm > min_overlap_threshold):
      return True
  return False


def _sample_positive_patches(image,
                             abnormality_filepath,
                             abnormality_mask,
                             abnormality_area,
                             patch_size,
                             number_of_patches=10,
                             min_overlap_threshold=0.90,
                             max_number_of_trials_per_threshold=100):
  """Sample random patches from the image overlapping with the given abnormality.

  The abnormal area of the patch with respect to either (a) the total area of
  the patch, or (b) the total area of the abnormality, must be at least
  `min_overlap_threshold` (i.e. 90% by default).

  After `max_number_of_trials_per_threshold` samples, if not enough patches
  meeting this requirement have been generated, the `min_overlap_threshold` is
  reduced by 5%. This procedure is repeated until min_overlap_threshold < 0.1
  (which should not happen ever, if the dataset is correct).

  Args:
    image: Image to patch from.
    abnormality_filepath: Only used for logging.
    abnormality_mask: Binary mask of the abnormality in the image.
    abnormality_area: Precomputed area of the abnormality.
    patch_size: Size of the patch to extract.
    number_of_patches: Number of patches to sample around the abnormality ROI.
    min_overlap_threshold: Minimum relative area of the patch overlapping with
      the abnormality.
    max_number_of_trials_per_threshold: Maximum number of random samples to try
      before reducing the `min_overlap_threshold` by 5%.

  Yields:
    The patch cropped from the input image.
  """
  cv2 = tfds.core.lazy_imports.cv2

  # The paper trying to be reproduced states that 90% of the are of each
  # positive patch should correspond to abnormal tissue. Thus if the total area
  # of abnormality is smaller than 0.9 * patch_area, we are certain that no
  # patch can meet this requirement. This happens indeed quite often.
  #
  # However, in a piece of code release by the authors of the paper
  # (https://github.com/yuyuyu123456/CBIS-DDSM/blob/bf3abc6ac2890b9b51eb5125e00056e39295fa44/ddsm_train/sample_patches_combined.py#L26)
  # the authors accept a patch if the total area of abnormality in the patch is
  # greater than 75% OR if 75% of the total abnormal area is in the patch.
  # In addition, they reduce the overlapping threholds every 1000 trials to
  # handle some corner casses.

  abnormality_roi = _get_roi_from_mask(abnormality_mask)
  abnorm_x, abnorm_y, abnorm_w, abnorm_h = cv2.boundingRect(abnormality_roi)

  number_of_yielded_patches = 0
  while min_overlap_threshold > 0.1:
    # Determine the region where random samples should be sampled from.
    max_h, min_h = max(abnorm_h, patch_size[0]), min(abnorm_h, patch_size[0])
    max_w, min_w = max(abnorm_w, patch_size[1]), min(abnorm_w, patch_size[1])
    min_y = abnorm_y - int((1.0 - min_overlap_threshold) * max_h)
    min_x = abnorm_x - int((1.0 - min_overlap_threshold) * max_w)
    max_y = abnorm_y + abnorm_h - int(min_overlap_threshold * min_h)
    max_x = abnorm_x + abnorm_w - int(min_overlap_threshold * min_w)
    # Ensure that all sampled batches are within the image.
    min_y = max(min_y, 0)
    min_x = max(min_x, 0)
    max_y = max(min(max_y, image.shape[0] - patch_size[0] - 1), min_y)
    max_x = max(min(max_x, image.shape[1] - patch_size[1] - 1), min_x)
    # Cap the number of trials if the sampling region is too small.
    effective_range_size = max_number_of_trials_per_threshold
    if (max_y - min_y + 1) * (max_x - min_x + 1) < effective_range_size:
      logging.debug(
          'The sampling region for patches of size %r with '
          'min_overlap_threshold=%f contains less possible patches than '
          'max_number_of_trials_per_threshold=%d, in abnormality %s',
          patch_size, min_overlap_threshold, max_number_of_trials_per_threshold,
          abnormality_filepath)
      effective_range_size = (max_y - min_y + 1) * (max_x - min_x + 1)

    for _ in range(effective_range_size):
      patch_y = np.random.randint(min_y, max_y + 1)
      patch_x = np.random.randint(min_x, max_x + 1)
      if _patch_overlaps_any_abnormality_above_threshold(
          patch_y, patch_x, patch_size, [abnormality_mask],
          [abnormality_area], min_overlap_threshold):
        number_of_yielded_patches += 1
        yield image[patch_y:(patch_y + patch_size[0]), patch_x:(patch_x +
                                                                patch_size[1])]
      # If we have yielded all requested patches return.
      if number_of_yielded_patches >= number_of_patches:
        return
    # We failed to produce patches with the minimum overlapping requirements.
    # Reduce those requirements and try again.
    min_overlap_threshold = min_overlap_threshold * 0.95
    logging.debug(
        'Overlapping constraints relaxed to min_overlap_threshold=%f while '
        'sampling positive patches for the abnormality %s',
        min_overlap_threshold, abnormality_filepath)

  # This should not happen ever.
  raise ValueError(
      'Only %d positive patches of size %r could be sampled satisfying the '
      'current conditions (min. relative overlapping area = %f) for the '
      'abnormality %s' % (number_of_yielded_patches, patch_size,
                          min_overlap_threshold, abnormality_filepath))


def _sample_negative_patches(image,
                             image_filepath,
                             abnormalities_masks,
                             abnormalities_areas,
                             patch_size,
                             number_of_patches=10,
                             min_breast_overlap_threshold=0.75,
                             max_abnorm_overlap_threshold=0.35,
                             max_number_of_trials_per_threshold=100):
  """Sample background patches from the image.

  The relative area of breast tissue in the patch must be, at least,
  `min_breast_overlap_threshold` of the total patch area. This is to prevent
  too easy negative examples.

  Similarly, the relative area of the abnormal tissue in the patch must be,
  at most, `max_abnorm_overlap_threshold`

  The relative area of the patch must overlap with the breast tissue with,
  at least, `min_breast_overlap_threshold` (relative) pixels.
  In addition, it must also overlap with abnormal tissue with, at most,
  `max_abnorm_overlap_threshold` (relative) pixels.

  Args:
    image: Image to patch from.
    image_filepath: Only used for logging.
    abnormalities_masks: List of binary mask of each abnormality in the image.
    abnormalities_areas: List of precomputed area of each abnormality.
    patch_size: Size of the patch to extract.
    number_of_patches: Number of negative patches to sample from the image.
    min_breast_overlap_threshold: Minimum (relative) number of breast pixels in
      the patch.
    max_abnorm_overlap_threshold: Maximum (relative) number of abnormal pixels
      in the patch.
    max_number_of_trials_per_threshold: Maximum number of random samples to try
      before reducing the `min_breast_overlap_threshold` by 5% and increasing
      the `max_abnorm_overlap_threshold` by 5%.

  Yields:
    The patch cropped from the input image.
  """
  cv2 = tfds.core.lazy_imports.cv2

  breast_mask = _get_breast_mask(image)

  def patch_overlapping_breast_is_feasible(y, x):
    """Return True if the patch contains enough breast pixels."""
    breast_in_patch = breast_mask[y:(y + patch_size[0]), x:(x + patch_size[1])]
    return (np.sum(breast_in_patch > 0) /
            (patch_size[0] * patch_size[1]) > min_breast_overlap_threshold)

  breast_roi = _get_roi_from_mask(breast_mask)
  breast_x, breast_y, breast_w, breast_h = cv2.boundingRect(breast_roi)
  number_of_yielded_patches = 0
  while (min_breast_overlap_threshold > 0.1 and
         max_abnorm_overlap_threshold < 0.9):
    # Determine the region where random samples should be sampled from.
    max_h, min_h = max(breast_h, patch_size[0]), min(breast_h, patch_size[0])
    max_w, min_w = max(breast_w, patch_size[1]), min(breast_w, patch_size[1])
    min_y = breast_y - int((1.0 - min_breast_overlap_threshold) * max_h)
    min_x = breast_x - int((1.0 - min_breast_overlap_threshold) * max_w)
    max_y = breast_y + breast_h - int(min_breast_overlap_threshold * min_h)
    max_x = breast_x + breast_w - int(min_breast_overlap_threshold * min_w)
    # Ensure that all sampled batches are within the image.
    min_y = max(min_y, 0)
    min_x = max(min_x, 0)
    max_y = max(min(max_y, image.shape[0] - patch_size[0] - 1), min_y)
    max_x = max(min(max_x, image.shape[1] - patch_size[1] - 1), min_x)
    # Cap the number of trials if the sampling region is too small.
    effective_range_size = max_number_of_trials_per_threshold
    if (max_y - min_y + 1) * (max_x - min_x + 1) < effective_range_size:
      logging.debug(
          'The sampling region for negative patches of size %r with '
          'min_breast_overlap_threshold=%f contains less possible patches '
          'than max_number_of_trials_per_threshold=%d, in mammography %s',
          patch_size, min_breast_overlap_threshold,
          max_number_of_trials_per_threshold, image_filepath)
      effective_range_size = (max_y - min_y + 1) * (max_x - min_x + 1)
    for _ in range(effective_range_size):
      patch_y = np.random.randint(min_y, max_y + 1)
      patch_x = np.random.randint(min_x, max_x + 1)
      if (patch_overlapping_breast_is_feasible(patch_y, patch_x) and
          not _patch_overlaps_any_abnormality_above_threshold(
              patch_y, patch_x, patch_size, abnormalities_masks,
              abnormalities_areas, max_abnorm_overlap_threshold)):
        number_of_yielded_patches += 1
        yield image[patch_y:(patch_y + patch_size[0]), patch_x:(patch_x +
                                                                patch_size[1])]
      # If we have yielded all requested patches return.
      if number_of_yielded_patches >= number_of_patches:
        return
    # We failed to produce patches with the given overlapping requirements.
    # Relaxate the requirements and try again.
    min_breast_overlap_threshold = min_breast_overlap_threshold * 0.95
    max_abnorm_overlap_threshold = max_abnorm_overlap_threshold * 1.05
    logging.debug(
        'Overlapping constraints relaxed to min_breast_overlap_threshold=%f '
        'and max_abnorm_overlap_threshold=%f while sampling negative '
        'patches for the mammography %s', min_breast_overlap_threshold,
        max_abnorm_overlap_threshold,
        image_filepath)  # Filepath to the abnormality mask image.

  # This should not happen ever.
  raise ValueError(
      'Only %d negative patches of size %r could be sampled satisfying the '
      'current conditions (min. relative overlapping area with breast = %f, '
      'max. relative overlapping area with abnormalities = %f) for the '
      'mammography %s' %
      (number_of_yielded_patches, patch_size, min_breast_overlap_threshold,
       max_abnorm_overlap_threshold, image_filepath))



# coding=utf-8
# Copyright 2019 The TensorFlow Datasets Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

r"""Generate ColorectalHistology-like files with random data."""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import os

from absl import app
from absl import flags
import numpy as np
import tensorflow as tf

from tensorflow_datasets.core import lazy_imports_lib
from tensorflow_datasets.core.utils import py_utils
from tensorflow_datasets.image import colorectal_histology

flags.DEFINE_string("tfds_dir", py_utils.tfds_dir(),
                    "Path to tensorflow_datasets directory")
# --compression=raw may be more portable, but results in massive files (>100mb)
flags.DEFINE_string(
    "compression", "tiff_lzw", "Used by PIL to compress fake images")
FLAGS = flags.FLAGS

# pylint: disable=protected-access

num_classes = len(colorectal_histology._CLASS_NAMES)
class_index = {c: i for i, c in enumerate(colorectal_histology._CLASS_NAMES)}


def examples_dir():
  return os.path.join(FLAGS.tfds_dir, "testing", "test_data", "fake_examples")


def histology_dir(large=False):
  folder = os.path.join(examples_dir(), "colorectal_histology")
  if large:
    folder = "%s_large" % folder
  return folder


def make_images(num_images, size):
  # random values compress badly
  return np.zeros((num_images, size, size, 3), dtype=np.uint8)


def write_image(filename, data):
  lazy_imports_lib.lazy_imports.PIL_Image.fromarray(data).save(
      filename, compression=FLAGS.compression)


def main(_):
  base_dir = os.path.join(
      histology_dir(False), colorectal_histology._TILES_SUBDIR)
  for ci, class_name in enumerate(colorectal_histology._CLASS_NAMES):
    subdir = os.path.join(
        base_dir, colorectal_histology._class_subdir(ci, class_name))
    tf.io.gfile.makedirs(subdir)

    for i, image_data in enumerate(
        make_images(2, colorectal_histology._TILES_SIZE)):
      fn = "image%d.tif" % i
      write_image(os.path.join(subdir, fn), image_data)

  base_dir = os.path.join(
      histology_dir(True), colorectal_histology._LARGE_SUBDIR)
  tf.io.gfile.makedirs(base_dir)
  write_image(
      os.path.join(base_dir, "large_image.tif"),
      make_images(1, colorectal_histology._LARGE_SIZE)[0])


if __name__ == "__main__":
  app.run(main)



# coding=utf-8
# Copyright 2019 The TensorFlow Datasets Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Stanford dogs dataset."""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import collections
import os
import re
import xml.etree.ElementTree as ET

import tensorflow as tf
import tensorflow_datasets.public_api as tfds

_DESCRIPTION = """\
The Stanford Dogs dataset contains images of 120 breeds of dogs from around
the world. This dataset has been built using images and annotation from
ImageNet for the task of fine-grained image categorization. There are
20,580 images, out of which 12,000 are used for training and 8580 for
testing. Class labels and bounding box annotations are provided
for all the 12,000 images.
"""

_URL = ("http://vision.stanford.edu/aditya86/ImageNetDogs/main.html")

_CITATION = """\
@inproceedings{KhoslaYaoJayadevaprakashFeiFei_FGVC2011,
author = "Aditya Khosla and Nityananda Jayadevaprakash and Bangpeng Yao and
          Li Fei-Fei",
title = "Novel Dataset for Fine-Grained Image Categorization",
booktitle = "First Workshop on Fine-Grained Visual Categorization,
             IEEE Conference on Computer Vision and Pattern Recognition",
year = "2011",
month = "June",
address = "Colorado Springs, CO",
}
@inproceedings{imagenet_cvpr09,
        AUTHOR = {Deng, J. and Dong, W. and Socher, R. and Li, L.-J. and
                  Li, K. and Fei-Fei, L.},
        TITLE = {{ImageNet: A Large-Scale Hierarchical Image Database}},
        BOOKTITLE = {CVPR09},
        YEAR = {2009},
        BIBSOURCE = "http://www.image-net.org/papers/imagenet_cvpr09.bib"}
"""

_IMAGES_URL = "http://vision.stanford.edu/aditya86/ImageNetDogs/images.tar"
_SPLIT_URL = "http://vision.stanford.edu/aditya86/ImageNetDogs/lists.tar"
_ANNOTATIONS_URL = "http://vision.stanford.edu/aditya86/ImageNetDogs/annotation.tar"
_NAME_RE = re.compile(r"([\w-]*/)*([\w]*.jpg)$")


class StanfordDogs(tfds.core.GeneratorBasedBuilder):
  """Stanford Dogs dataset."""

  VERSION = tfds.core.Version("0.1.0")

  def _info(self):

    return tfds.core.DatasetInfo(
        builder=self,
        description=_DESCRIPTION,
        features=tfds.features.FeaturesDict({
            # Images are of varying size
            "image":
                tfds.features.Image(),
            "image/filename":
                tfds.features.Text(),
            "label":
                tfds.features.ClassLabel(num_classes=120),
            # Multiple bounding box per image
            "objects":
                tfds.features.Sequence({
                    "bbox": tfds.features.BBoxFeature(),
                }),
        }),
        supervised_keys=("image", "label"),
        urls=[_URL],
        citation=_CITATION)

  def _split_generators(self, dl_manager):

    download_path = dl_manager.download(
        [_SPLIT_URL, _ANNOTATIONS_URL, _IMAGES_URL])
    extracted_path = dl_manager.download_and_extract(
        [_SPLIT_URL, _ANNOTATIONS_URL])
    xml_file_list = collections.defaultdict(str)

    # Parsing the mat file which contains the list of train/test images
    def parse_mat_file(file_name):

      parsed_mat_arr = tfds.core.lazy_imports.scipy.io.loadmat(
          file_name, squeeze_me=True)
      file_list = [
          element.split("/")[-1] for element in parsed_mat_arr["file_list"]
      ]

      return file_list, parsed_mat_arr

    for fname in tf.io.gfile.listdir(extracted_path[0]):
      # Train-test split using train_list.mat and test_list.mat
      full_file_name = os.path.join(extracted_path[0], fname)

      if "train" in fname:
        train_list, train_mat_arr = parse_mat_file(full_file_name)
        label_names = set([
            element.split("/")[-2].lower()
            for element in train_mat_arr["file_list"]
        ])
      elif "test" in fname:
        test_list, _ = parse_mat_file(full_file_name)

    self.info.features["label"].names = tuple(label_names)

    for root, _, files in tf.io.gfile.walk(extracted_path[1]):
      # Parsing the XML file which have the image annotations
      for fname in files:
        annotation_file_name = os.path.join(root, fname)
        xml_file_list[fname] = ET.parse(annotation_file_name)

    return [
        tfds.core.SplitGenerator(
            name=tfds.Split.TRAIN,
            num_shards=10,
            gen_kwargs={
                "archive": dl_manager.iter_archive(download_path[2]),
                "file_names": train_list,
                "annotation_files": xml_file_list,
            }),
        tfds.core.SplitGenerator(
            name=tfds.Split.TEST,
            num_shards=1,
            gen_kwargs={
                "archive": dl_manager.iter_archive(download_path[2]),
                "file_names": test_list,
                "annotation_files": xml_file_list,
            })
    ]

  def _generate_examples(self, archive, file_names, annotation_files):
    """Generate dog images, labels, bbox attributes given the directory path.

    Args:
      archive: object that iterates over the zip
      file_names : list of train/test image file names obtained from mat file
      annotation_files : dict of image file names and their xml object

    Yields:
      Image path, Image file name, its corresponding label and
      bounding box values
    """
    label_names = self.info.features["label"].names
    bbox_attrib = ["xmin", "xmax", "ymin", "ymax", "width", "height"]

    for fname, fobj in archive:
      res = _NAME_RE.match(fname)
      if not res or (fname.split("/")[-1] not in file_names):
        continue

      label = res.group(1)[:-1].lower()
      file_name = res.group(2)
      attributes = collections.defaultdict(list)
      for element in annotation_files[file_name.split(".")[0]].iter():
        # Extract necessary Bbox attributes from XML file
        if element.tag.strip() in bbox_attrib:
          attributes[element.tag.strip()].append(float(element.text.strip()))

      # BBox attributes in range of 0.0 to 1.0
      def normalize_bbox(bbox_side, image_side):
        return min(bbox_side / image_side, 1.0)

      def build_box(attributes, n):
        return tfds.features.BBox(
            ymin=normalize_bbox(attributes["ymin"][n], attributes["height"][0]),
            xmin=normalize_bbox(attributes["xmin"][n], attributes["width"][0]),
            ymax=normalize_bbox(attributes["ymax"][n], attributes["height"][0]),
            xmax=normalize_bbox(attributes["xmax"][n], attributes["width"][0]),
        )

      yield fname, {
          "image":
              fobj,
          "image/filename":
              fname,
          "label":
              label_names.index(label),
          "objects": [{
              "bbox": build_box(attributes, n)
          } for n in range(len(attributes["xmin"]))]
      }



# coding=utf-8
# Copyright 2019 The TensorFlow Datasets Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Video feature."""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import os
import subprocess
import tempfile

import six
import tensorflow as tf
from tensorflow_datasets.core.features import image_feature
from tensorflow_datasets.core.features import sequence_feature


class Video(sequence_feature.Sequence):
  """`FeatureConnector` for videos, encoding frames individually on disk.

  Video: The image connector accepts as input a 4 dimensional uint8 array
  representing a video, a sequence of paths to encoded frames, or a path or a
  file object that can be decoded with ffmpeg. Note that not all formats in
  ffmpeg support reading from pipes, so providing a file object might fail.
  Furthermore, if a path is given that is not on the local file system, we first
  copy it to a temporary local file before passing it to ffmpeg.

  Output:
    video: tf.Tensor of type tf.uint8 and shape
      [num_frames, height, width, channels], where channels must be 1 or 3

  Example:
    * In the DatasetInfo object:
        features=features.FeatureDict({
            'video': features.Video(shape=(None, 64, 64, 3)),
        })

    * During generation:
        ```
        yield {
            'input': np.ones(shape=(128, 64, 64, 3), dtype=np.uint8),
        }
        ```
        or
        ```
        yield {
    '      video': ['path/to/frame001.png', 'path/to/frame002.png'],
        }
        ```
        or
        ```
        yield {
              'input': '/path/to/video.avi',
        }
        ```
        or
        ```
        yield {
              'input': gfile.GFile('/complex/path/video.avi'),
        }
        ```
  """

  def __init__(self, shape, encoding_format='png', ffmpeg_extra_args=()):
    """Initializes the connector.

    Args:
      shape: tuple of ints, the shape of the video (num_frames, height, width,
        channels), where channels is 1 or 3.
      encoding_format: The video is stored as a sequence of encoded images.
        You can use any encoding format supported by image_feature.Feature.
      ffmpeg_extra_args: A sequence of additional args to be passed to the
        ffmpeg binary. Specifically, ffmpeg will be called as:
          ``
          ffmpeg -i <input_file> <ffmpeg_extra_args> %010d.<encoding_format>
          ``
    Raises:
      ValueError: If the shape is invalid
    """
    shape = tuple(shape)
    if len(shape) != 4:
      raise ValueError('Video shape should be of rank 4')
    self._encoding_format = encoding_format
    self._extra_ffmpeg_args = list(ffmpeg_extra_args or [])
    super(Video, self).__init__(
        image_feature.Image(shape=shape[1:], encoding_format=encoding_format),
        length=shape[0],
    )

  @property
  def _ffmpeg_path(self):
    return 'ffmpeg'


  def _ffmpeg_decode(self, path_or_fobj):
    if isinstance(path_or_fobj, six.string_types):
      ffmpeg_args = [self._ffmpeg_path, '-i', path_or_fobj]
      ffmpeg_stdin = None
    else:
      ffmpeg_args = [self._ffmpeg_path, '-i', 'pipe:0']
      ffmpeg_stdin = path_or_fobj.read()

    ffmpeg_dir = tempfile.mkdtemp()
    output_pattern = os.path.join(ffmpeg_dir, '%010d.' + self._encoding_format)
    ffmpeg_args += self._extra_ffmpeg_args
    ffmpeg_args.append(output_pattern)
    try:
      process = subprocess.Popen(ffmpeg_args,
                                 stdin=subprocess.PIPE,
                                 stdout=subprocess.PIPE,
                                 stderr=subprocess.PIPE)
      stdout_data, stderr_data = process.communicate(ffmpeg_stdin)
      ffmpeg_ret_code = process.returncode
      if ffmpeg_ret_code:
        raise ValueError(
            'ffmpeg returned error code {}, command={}\n'
            'stdout={}\nstderr={}\n'.format(ffmpeg_ret_code,
                                            ' '.join(ffmpeg_args),
                                            stdout_data,
                                            stderr_data))
      frames = []
      for image_name in sorted(tf.io.gfile.listdir(ffmpeg_dir)):
        image_path = os.path.join(ffmpeg_dir, image_name)
        with tf.io.gfile.GFile(image_path, 'rb') as frame_file:
          frames.append(six.BytesIO(frame_file.read()))
      return frames
    except OSError as exception:
      raise IOError(
          'It seems that ffmpeg is not installed on the system. Please follow '
          'the instrutions at https://ffmpeg.org/. '
          'Original exception: {}'.format(exception))
    finally:
      tf.io.gfile.rmtree(ffmpeg_dir)

  def encode_example(self, video_or_path_or_fobj):
    """Converts the given image into a dict convertible to tf example."""
    if isinstance(video_or_path_or_fobj, six.string_types):
      if not os.path.isfile(video_or_path_or_fobj):
        _, video_temp_path = tempfile.mkstemp()
        try:
          tf.gfile.Copy(video_or_path_or_fobj, video_temp_path, overwrite=True)
          encoded_video = self._ffmpeg_decode(video_temp_path)
        finally:
          os.unlink(video_temp_path)
      else:
        encoded_video = self._ffmpeg_decode(video_or_path_or_fobj)
    elif hasattr(video_or_path_or_fobj, 'read'):
      encoded_video = self._ffmpeg_decode(video_or_path_or_fobj)
    else:
      encoded_video = video_or_path_or_fobj
    return super(Video, self).encode_example(encoded_video)



# coding=utf-8
# Copyright 2019 The TensorFlow Datasets Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Translation datasets."""

from tensorflow_datasets.translate.flores import Flores
from tensorflow_datasets.translate.flores import FloresConfig
from tensorflow_datasets.translate.para_crawl import ParaCrawl
from tensorflow_datasets.translate.para_crawl import ParaCrawlConfig
from tensorflow_datasets.translate.ted_hrlr import TedHrlrTranslate
from tensorflow_datasets.translate.ted_multi import TedMultiTranslate
from tensorflow_datasets.translate.wmt import WmtConfig
from tensorflow_datasets.translate.wmt14 import Wmt14Translate
from tensorflow_datasets.translate.wmt15 import Wmt15Translate
from tensorflow_datasets.translate.wmt16 import Wmt16Translate
from tensorflow_datasets.translate.wmt17 import Wmt17Translate
from tensorflow_datasets.translate.wmt18 import Wmt18Translate
from tensorflow_datasets.translate.wmt19 import Wmt19Translate
from tensorflow_datasets.translate.wmt_t2t import WmtT2tTranslate



# coding=utf-8
# Copyright 2019 The TensorFlow Datasets Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests for the GLUE data set.

We have an individual test for each config so that we can use sharding to
prevent the test from timing out.
"""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

from tensorflow_datasets import testing
from tensorflow_datasets.text import glue


class GlueColaTest(testing.DatasetBuilderTestCase):
  BUILDER_CONFIG_NAMES_TO_TEST = ["cola"]
  DATASET_CLASS = glue.Glue
  SPLITS = {
      "train": 3,
      "validation": 2,
      "test": 1,
  }


class GlueSst2Test(testing.DatasetBuilderTestCase):
  BUILDER_CONFIG_NAMES_TO_TEST = ["sst2"]
  DATASET_CLASS = glue.Glue
  SPLITS = {
      "train": 3,
      "validation": 2,
      "test": 1,
  }


class GlueQqpTest(testing.DatasetBuilderTestCase):
  BUILDER_CONFIG_NAMES_TO_TEST = ["qqp"]
  DATASET_CLASS = glue.Glue
  SPLITS = {
      "train": 3,
      "validation": 2,
      "test": 1,
  }


class GlueStsbTest(testing.DatasetBuilderTestCase):
  BUILDER_CONFIG_NAMES_TO_TEST = ["stsb"]
  DATASET_CLASS = glue.Glue
  SPLITS = {
      "train": 3,
      "validation": 2,
      "test": 1,
  }


class GlueMnliTest(testing.DatasetBuilderTestCase):
  BUILDER_CONFIG_NAMES_TO_TEST = ["mnli"]
  DATASET_CLASS = glue.Glue
  SPLITS = {
      "train": 3,
      "validation_matched": 2,
      "validation_mismatched": 2,
      "test_matched": 1,
      "test_mismatched": 1,
  }


class GlueQnliTest(testing.DatasetBuilderTestCase):
  BUILDER_CONFIG_NAMES_TO_TEST = ["qnli"]
  DATASET_CLASS = glue.Glue
  SPLITS = {
      "train": 3,
      "validation": 2,
      "test": 1,
  }


class GlueRteTest(testing.DatasetBuilderTestCase):
  BUILDER_CONFIG_NAMES_TO_TEST = ["rte"]
  DATASET_CLASS = glue.Glue
  SPLITS = {
      "train": 3,
      "validation": 2,
      "test": 1,
  }


class GlueWnliTest(testing.DatasetBuilderTestCase):
  BUILDER_CONFIG_NAMES_TO_TEST = ["wnli"]
  DATASET_CLASS = glue.Glue
  SPLITS = {
      "train": 3,
      "validation": 2,
      "test": 1,
  }


class GlueMrpcTest(testing.DatasetBuilderTestCase):
  BUILDER_CONFIG_NAMES_TO_TEST = ["mrpc"]
  DATASET_CLASS = glue.Glue
  DL_EXTRACT_RESULT = {
      "train": "MRPC/msr_paraphrase_train.txt",
      "test": "MRPC/msr_paraphrase_test.txt",
      "dev_ids": "MRPC/mrpc_dev_ids.tsv",
  }
  SPLITS = {
      "train": 10,
      "validation": 8,
      "test": 15,
  }


class GlueColaS3Test(GlueColaTest):
  VERSION = "experimental_latest"


class GlueSst2S3Test(GlueSst2Test):
  VERSION = "experimental_latest"


class GlueQqpS3Test(GlueQqpTest):
  VERSION = "experimental_latest"


class GlueStsbS3Test(GlueStsbTest):
  VERSION = "experimental_latest"


class GlueMnliS3Test(GlueMnliTest):
  VERSION = "experimental_latest"


class GlueQnliS3Test(GlueQnliTest):
  VERSION = "experimental_latest"


class GlueRteS3Test(GlueRteTest):
  VERSION = "experimental_latest"


class GlueWnliS3Test(GlueWnliTest):
  VERSION = "experimental_latest"


class GlueMrpcS3Test(GlueMrpcTest):
  VERSION = "experimental_latest"


if __name__ == "__main__":
  testing.test_main()



# coding=utf-8
# Copyright 2019 The TensorFlow Datasets Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Text feature.

"""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import os

import tensorflow as tf

from tensorflow_datasets.core.features import feature
from tensorflow_datasets.core.features import text as text_lib


class Text(feature.Tensor):
  """`FeatureConnector` for text, encoding to integers with a `TextEncoder`."""

  def __init__(self, encoder=None, encoder_config=None):
    """Constructs a Text FeatureConnector.

    Args:
      encoder: `tfds.features.text.TextEncoder`, an encoder that can convert
        text to integers. If None, the text will be utf-8 byte-encoded.
      encoder_config: `tfds.features.text.TextEncoderConfig`, needed if
        restoring from a file with `load_metadata`.
    """
    if encoder and encoder_config:
      raise ValueError("If encoder is provided, encoder_config must be None.")
    if encoder:
      encoder_config = text_lib.TextEncoderConfig(
          encoder_cls=type(encoder),
          vocab_size=encoder.vocab_size)
    elif encoder_config:
      encoder = encoder_config.encoder

    self._encoder = encoder
    self._encoder_config = encoder_config

    has_encoder = bool(encoder or self._encoder_cls)
    super(Text, self).__init__(
        shape=(None,) if has_encoder else (),
        dtype=tf.int64 if has_encoder else tf.string,
    )

  @property
  def encoder(self):
    return self._encoder

  @encoder.setter
  def encoder(self, new_encoder):
    if self.encoder:
      raise ValueError("Cannot override encoder")
    self._encoder = new_encoder
    encoder_cls = self._encoder_cls or type(None)
    if not isinstance(new_encoder, encoder_cls):
      raise ValueError(
          "Changing type of encoder. Got %s but must be %s" %
          (type(new_encoder).__name__,
           self._encoder_cls.__name__))

  def maybe_set_encoder(self, new_encoder):
    """Set encoder, but no-op if encoder is already set."""
    if self.encoder:
      return
    self.encoder = new_encoder

  @property
  def vocab_size(self):
    return self.encoder and self.encoder.vocab_size

  def str2ints(self, str_value):
    """Conversion string => encoded list[int]."""
    if not self._encoder:
      raise ValueError(
          "Text.str2ints is not available because encoder hasn't been defined.")
    return self._encoder.encode(str_value)

  def ints2str(self, int_values):
    """Conversion list[int] => decoded string."""
    if not self._encoder:
      raise ValueError(
          "Text.ints2str is not available because encoder hasn't been defined.")
    return self._encoder.decode(int_values)

  def encode_example(self, example_data):
    if self.encoder:
      example_data = self.encoder.encode(example_data)
    else:
      example_data = example_data
    return super(Text, self).encode_example(example_data)

  def save_metadata(self, data_dir, feature_name):
    fname_prefix = os.path.join(data_dir, "%s.text" % feature_name)
    if not self.encoder:
      return
    self.encoder.save_to_file(fname_prefix)

  def load_metadata(self, data_dir, feature_name):
    fname_prefix = os.path.join(data_dir, "%s.text" % feature_name)
    encoder_cls = self._encoder_cls
    if encoder_cls:
      self._encoder = encoder_cls.load_from_file(fname_prefix)
      return

    # Error checking: ensure there are no metadata files
    feature_files = [
        f for f in tf.io.gfile.listdir(data_dir) if f.startswith(fname_prefix)
    ]
    if feature_files:
      raise ValueError(
          "Text feature files found for feature %s but encoder_cls=None. "
          "Make sure to set encoder_cls in the TextEncoderConfig. "
          "Files: %s" % (feature_name, feature_files))

  def maybe_build_from_corpus(self, corpus_generator, **kwargs):
    """Call SubwordTextEncoder.build_from_corpus is encoder_cls is such.

    If `self.encoder` is `None` and `self._encoder_cls` is of type
    `SubwordTextEncoder`, the method instantiates `self.encoder` as returned
    by `SubwordTextEncoder.build_from_corpus()`.

    Args:
      corpus_generator: generator yielding `str`, from which
        subwords will be constructed.
      **kwargs: kwargs forwarded to `SubwordTextEncoder.build_from_corpus()`
    """
    if self._encoder_cls is not text_lib.SubwordTextEncoder:
      return
    if self.encoder:
      return

    vocab_size = self._encoder_config.vocab_size
    self.encoder = text_lib.SubwordTextEncoder.build_from_corpus(
        corpus_generator=corpus_generator,
        target_vocab_size=vocab_size,
        **kwargs)

  @property
  def _encoder_cls(self):
    return self._encoder_config and self._encoder_config.encoder_cls

  def _additional_repr_info(self):
    if self.encoder is None:
      return {}
    return {"encoder": repr(self.encoder)}



# coding=utf-8
# Copyright 2019 The TensorFlow Datasets Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests for tensorflow_datasets.core.dataset_info."""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import json
import os
import tempfile
import numpy as np
import six
import tensorflow as tf
from tensorflow_datasets import testing
from tensorflow_datasets.core import dataset_info
from tensorflow_datasets.core import features
from tensorflow_datasets.core.utils import py_utils
from tensorflow_datasets.image import mnist

tf.compat.v1.enable_eager_execution()

_TFDS_DIR = py_utils.tfds_dir()
_INFO_DIR = os.path.join(_TFDS_DIR, "testing", "test_data", "dataset_info",
                         "mnist", "1.0.0")
_INFO_DIR_UNLABELED = os.path.join(_TFDS_DIR, "testing", "test_data",
                                   "dataset_info", "mnist_unlabeled", "1.0.0")
_NON_EXISTENT_DIR = os.path.join(_TFDS_DIR, "non_existent_dir")


DummyDatasetSharedGenerator = testing.DummyDatasetSharedGenerator


class RandomShapedImageGenerator(DummyDatasetSharedGenerator):

  def _info(self):
    return dataset_info.DatasetInfo(
        builder=self,
        features=features.FeaturesDict({"im": features.Image()}),
        supervised_keys=("im", "im"),
        metadata=dataset_info.MetadataDict(),
    )

  def _generate_examples(self, range_):
    self.info.metadata["some_key"] = 123

    for i in range_:
      height = np.random.randint(5, high=10)
      width = np.random.randint(5, high=10)
      yield i, {
          "im":
              np.random.randint(
                  0, 255, size=(height, width, 3), dtype=np.uint8)
      }


class DatasetInfoTest(testing.TestCase):

  @classmethod
  def setUpClass(cls):
    super(DatasetInfoTest, cls).setUpClass()
    cls._tfds_tmp_dir = testing.make_tmp_dir()
    cls._builder = DummyDatasetSharedGenerator(data_dir=cls._tfds_tmp_dir)

  @classmethod
  def tearDownClass(cls):
    super(DatasetInfoTest, cls).tearDownClass()
    testing.rm_tmp_dir(cls._tfds_tmp_dir)

  def test_undefined_dir(self):
    with self.assertRaisesWithPredicateMatch(ValueError,
                                             "undefined dataset_info_dir"):
      info = dataset_info.DatasetInfo(builder=self._builder)
      info.read_from_directory(None)

  def test_non_existent_dir(self):
    info = dataset_info.DatasetInfo(builder=self._builder)
    with self.assertRaisesWithPredicateMatch(
        tf.errors.NotFoundError, "No such file or dir"):
      info.read_from_directory(_NON_EXISTENT_DIR)

  def test_reading(self):
    info = dataset_info.DatasetInfo(builder=self._builder)
    info.read_from_directory(_INFO_DIR)

    # Assert that we read the file and initialized DatasetInfo.
    self.assertTrue(info.initialized)
    self.assertEqual("dummy_dataset_shared_generator", info.name)
    self.assertEqual("dummy_dataset_shared_generator/1.0.0", info.full_name)

    # Test splits are initialized properly.
    split_dict = info.splits

    # Assert they are the correct number.
    self.assertTrue(len(split_dict), 2)

    # Assert on what they are
    self.assertIn("train", split_dict)
    self.assertIn("test", split_dict)

    # Assert that this is computed correctly.
    self.assertEqual(40, info.splits.total_num_examples)

    self.assertEqual("image", info.supervised_keys[0])
    self.assertEqual("label", info.supervised_keys[1])

  def test_reading_empty_properties(self):
    info = dataset_info.DatasetInfo(builder=self._builder)
    info.read_from_directory(_INFO_DIR_UNLABELED)

    # Assert supervised_keys has not been set
    self.assertEqual(None, info.supervised_keys)

  def test_writing(self):
    # First read in stuff.
    mnist_builder = mnist.MNIST(
        data_dir=tempfile.mkdtemp(dir=self.get_temp_dir()))

    info = dataset_info.DatasetInfo(
        builder=mnist_builder, features=mnist_builder.info.features)
    info.read_from_directory(_INFO_DIR)

    # Read the json file into a string.
    with tf.io.gfile.GFile(info._dataset_info_path(_INFO_DIR)) as f:
      existing_json = json.load(f)

    # Now write to a temp directory.
    with testing.tmp_dir(self.get_temp_dir()) as tmp_dir:
      info.write_to_directory(tmp_dir)

      # Read the newly written json file into a string.
      with tf.io.gfile.GFile(info._dataset_info_path(tmp_dir)) as f:
        new_json = json.load(f)

      # Read the newly written LICENSE file into a string.
      with tf.io.gfile.GFile(info._license_path(tmp_dir)) as f:
        license_ = f.read()

    # Assert what was read and then written and read again is the same.
    self.assertEqual(existing_json, new_json)

    # Assert correct license was written.
    self.assertEqual(existing_json["redistributionInfo"]["license"], license_)

    if six.PY3:
      # Only test on Python 3 to avoid u'' formatting issues
      self.assertEqual(repr(info), INFO_STR)

  def test_restore_after_modification(self):
    # Create a DatasetInfo
    info = dataset_info.DatasetInfo(
        builder=self._builder,
        description="A description",
        supervised_keys=("input", "output"),
        urls=["some location"],
        citation="some citation",
        redistribution_info={"license": "some license"}
    )
    info.size_in_bytes = 456
    info.as_proto.schema.feature.add()
    info.as_proto.schema.feature.add()  # Add dynamic statistics
    info.download_checksums = {
        "url1": "some checksum",
        "url2": "some other checksum",
    }

    with testing.tmp_dir(self.get_temp_dir()) as tmp_dir:
      # Save it
      info.write_to_directory(tmp_dir)

      # If fields are not defined, then everything is restored from disk
      restored_info = dataset_info.DatasetInfo(builder=self._builder)
      restored_info.read_from_directory(tmp_dir)
      self.assertEqual(info.as_proto, restored_info.as_proto)

    with testing.tmp_dir(self.get_temp_dir()) as tmp_dir:
      # Save it
      info.write_to_directory(tmp_dir)

      # If fields are defined, then the code version is kept
      restored_info = dataset_info.DatasetInfo(
          builder=self._builder,
          supervised_keys=("input (new)", "output (new)"),
          urls=["some location (new)"],
          citation="some citation (new)",
          redistribution_info={"license": "some license (new)"}
      )
      restored_info.size_in_bytes = 789
      restored_info.as_proto.schema.feature.add()
      restored_info.as_proto.schema.feature.add()
      restored_info.as_proto.schema.feature.add()
      restored_info.as_proto.schema.feature.add()  # Add dynamic statistics
      restored_info.download_checksums = {
          "url2": "some other checksum (new)",
          "url3": "some checksum (new)",
      }

      restored_info.read_from_directory(tmp_dir)

      # Even though restored_info has been restored, informations defined in
      # the code overwrite informations from the json file.
      self.assertEqual(restored_info.description, "A description")
      self.assertEqual(
          restored_info.supervised_keys, ("input (new)", "output (new)"))
      self.assertEqual(restored_info.urls, ["some location (new)"])
      self.assertEqual(restored_info.citation, "some citation (new)")
      self.assertEqual(restored_info.redistribution_info.license,
                       "some license (new)")
      self.assertEqual(restored_info.size_in_bytes, 789)
      self.assertEqual(len(restored_info.as_proto.schema.feature), 4)
      self.assertEqual(restored_info.download_checksums, {
          "url2": "some other checksum (new)",
          "url3": "some checksum (new)",
      })

  def test_reading_from_gcs_bucket(self):
    # The base TestCase prevents GCS access, so we explicitly ask it to restore
    # access here.
    with self.gcs_access():
      mnist_builder = mnist.MNIST(
          data_dir=tempfile.mkdtemp(dir=self.get_temp_dir()))
      info = dataset_info.DatasetInfo(builder=mnist_builder)
      info = mnist_builder.info

      # A nominal check to see if we read it.
      self.assertTrue(info.initialized)
      self.assertEqual(10000, info.splits["test"].num_examples)

  def test_str_smoke(self):
    info = mnist.MNIST(data_dir="/tmp/some_dummy_dir").info
    _ = str(info)

  @testing.run_in_graph_and_eager_modes()
  def test_statistics_generation(self):
    with testing.tmp_dir(self.get_temp_dir()) as tmp_dir:
      builder = DummyDatasetSharedGenerator(data_dir=tmp_dir)
      builder.download_and_prepare()

      # Overall
      self.assertEqual(30, builder.info.splits.total_num_examples)

      # Per split.
      test_split = builder.info.splits["test"].get_proto()
      train_split = builder.info.splits["train"].get_proto()
      self.assertEqual(10, test_split.statistics.num_examples)
      self.assertEqual(20, train_split.statistics.num_examples)

  @testing.run_in_graph_and_eager_modes()
  def test_statistics_generation_variable_sizes(self):
    with testing.tmp_dir(self.get_temp_dir()) as tmp_dir:
      builder = RandomShapedImageGenerator(data_dir=tmp_dir)
      builder.download_and_prepare()

      # Get the expected type of the feature.
      schema_feature = builder.info.as_proto.schema.feature[0]
      self.assertEqual("im", schema_feature.name)

      self.assertEqual(-1, schema_feature.shape.dim[0].size)
      self.assertEqual(-1, schema_feature.shape.dim[1].size)
      self.assertEqual(3, schema_feature.shape.dim[2].size)

  def test_metadata(self):
    with testing.tmp_dir(self.get_temp_dir()) as tmp_dir:
      builder = RandomShapedImageGenerator(data_dir=tmp_dir)
      builder.download_and_prepare()
      # Metadata should have been created
      self.assertEqual(builder.info.metadata, {"some_key": 123})

      # Metadata should have been restored
      builder2 = RandomShapedImageGenerator(data_dir=tmp_dir)
      self.assertEqual(builder2.info.metadata, {"some_key": 123})

  def test_updates_on_bucket_info(self):

    info = dataset_info.DatasetInfo(builder=self._builder,
                                    description="won't be updated")
    # No statistics in the above.
    self.assertEqual(0, info.splits.total_num_examples)
    self.assertEqual(0, len(info.as_proto.schema.feature))

    # Partial update will happen here.
    info.read_from_directory(_INFO_DIR)

    # Assert that description (things specified in the code) didn't change
    # but statistics are updated.
    self.assertEqual("won't be updated", info.description)

    # These are dynamically computed, so will be updated.
    self.assertEqual(40, info.splits.total_num_examples)
    self.assertEqual(2, len(info.as_proto.schema.feature))


INFO_STR = """tfds.core.DatasetInfo(
    name='mnist',
    version=1.0.0,
    description='The MNIST database of handwritten digits.',
    urls=['https://storage.googleapis.com/cvdf-datasets/mnist/'],
    features=FeaturesDict({
        'image': Image(shape=(28, 28, 1), dtype=tf.uint8),
        'label': ClassLabel(shape=(), dtype=tf.int64, num_classes=10),
    }),
    total_num_examples=40,
    splits={
        'test': 20,
        'train': 20,
    },
    supervised_keys=('image', 'label'),
    citation=\"\"\"@article{lecun2010mnist,
      title={MNIST handwritten digit database},
      author={LeCun, Yann and Cortes, Corinna and Burges, CJ},
      journal={ATT Labs [Online]. Available: http://yann. lecun. com/exdb/mnist},
      volume={2},
      year={2010}
    }\"\"\",
    redistribution_info=license: "test license",
)
"""


if __name__ == "__main__":
  testing.test_main()



# coding=utf-8
# Copyright 2019 The TensorFlow Datasets Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""dSprites dataset test."""
from tensorflow_datasets.image import dsprites
import tensorflow_datasets.testing as tfds_test


class DspritesTest(tfds_test.DatasetBuilderTestCase):
  DATASET_CLASS = dsprites.Dsprites
  SPLITS = {"train": 5}
  DL_EXTRACT_RESULT = "dsprites_ndarray_co1sh3sc6or40x32y32_64x64.hdf5"


class DspritesS3Test(DspritesTest):
  VERSION = "experimental_latest"


if __name__ == "__main__":
  tfds_test.test_main()



# coding=utf-8
# Copyright 2019 The TensorFlow Datasets Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Download manager interface."""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import os
import sys
import uuid

from absl import logging
import promise
import six
import tensorflow as tf

from tensorflow_datasets.core import api_utils
from tensorflow_datasets.core import utils
from tensorflow_datasets.core.download import checksums
from tensorflow_datasets.core.download import downloader
from tensorflow_datasets.core.download import extractor
from tensorflow_datasets.core.download import resource as resource_lib
from tensorflow_datasets.core.download import util


class NonMatchingChecksumError(Exception):
  """The downloaded file doesn't have expected checksum."""

  def __init__(self, url, tmp_path):
    msg = 'Artifact %s, downloaded to %s, has wrong checksum.' % (url, tmp_path)
    Exception.__init__(self, msg)


class DownloadConfig(object):
  """Configuration for `tfds.core.DatasetBuilder.download_and_prepare`."""

  def __init__(self,
               extract_dir=None,
               manual_dir=None,
               download_mode=None,
               compute_stats=None,
               max_examples_per_split=None,
               register_checksums=False,
               beam_runner=None,
               beam_options=None):
    """Constructs a `DownloadConfig`.

    Args:
      extract_dir: `str`, directory where extracted files are stored.
        Defaults to "<download_dir>/extracted".
      manual_dir: `str`, read-only directory where manually downloaded/extracted
        data is stored. Defaults to
        "<download_dir>/manual".
      download_mode: `tfds.GenerateMode`, how to deal with downloads or data
        that already exists. Defaults to `REUSE_DATASET_IF_EXISTS`, which will
        reuse both downloads and data if it already exists.
      compute_stats: `tfds.download.ComputeStats`, whether to compute
        statistics over the generated data. Defaults to `AUTO`.
      max_examples_per_split: `int`, optional max number of examples to write
        into each split.
      register_checksums: `bool`, defaults to False. If True, checksum of
        downloaded files are recorded.
      beam_runner: Runner to pass to `beam.Pipeline`, only used for datasets
        based on Beam for the generation.
      beam_options: `PipelineOptions` to pass to `beam.Pipeline`, only used for
        datasets based on Beam for the generation.
    """
    self.extract_dir = extract_dir
    self.manual_dir = manual_dir
    self.download_mode = util.GenerateMode(
        download_mode or util.GenerateMode.REUSE_DATASET_IF_EXISTS)
    self.compute_stats = util.ComputeStatsMode(
        compute_stats or util.ComputeStatsMode.AUTO)
    self.max_examples_per_split = max_examples_per_split
    self.register_checksums = register_checksums
    self.beam_runner = beam_runner
    self.beam_options = beam_options


class DownloadManager(object):
  """Manages the download and extraction of files, as well as caching.

  Downloaded files are cached under `download_dir`. The file name of downloaded
   files follows pattern "${sanitized_url}${content_checksum}.${ext}". Eg:
   'cs.toronto.edu_kriz_cifar-100-pythonJDF[...]I.tar.gz'.

  While a file is being downloaded, it is placed into a directory following a
  similar but different pattern:
  "%{sanitized_url}${url_checksum}.tmp.${uuid}".

  When a file is downloaded, a "%{fname}s.INFO.json" file is created next to it.
  This INFO file contains the following information:
  {"dataset_names": ["name1", "name2"],
   "urls": ["http://url.of/downloaded_file"]}

  Extracted files/dirs are stored under `extract_dir`. The file name or
  directory name is the same as the original name, prefixed with the extraction
  method. E.g.
   "${extract_dir}/TAR_GZ.cs.toronto.edu_kriz_cifar-100-pythonJDF[...]I.tar.gz".

  The function members accept either plain value, or values wrapped into list
  or dict. Giving a data structure will parallelize the downloads.

  Example of usage:

  ```
  # Sequential download: str -> str
  train_dir = dl_manager.download_and_extract('https://abc.org/train.tar.gz')
  test_dir = dl_manager.download_and_extract('https://abc.org/test.tar.gz')

  # Parallel download: list -> list
  image_files = dl_manager.download(
      ['https://a.org/1.jpg', 'https://a.org/2.jpg', ...])

  # Parallel download: dict -> dict
  data_dirs = dl_manager.download_and_extract({
     'train': 'https://abc.org/train.zip',
     'test': 'https://abc.org/test.zip',
  })
  data_dirs['train']
  data_dirs['test']
  ```

  For more customization on the download/extraction (ex: passwords, output_name,
  ...), you can pass a `tfds.download.Resource` as argument.
  """

  @api_utils.disallow_positional_args
  def __init__(self,
               download_dir,
               extract_dir=None,
               manual_dir=None,
               dataset_name=None,
               force_download=False,
               force_extraction=False,
               register_checksums=False):
    """Download manager constructor.

    Args:
      download_dir: `str`, path to directory where downloads are stored.
      extract_dir: `str`, path to directory where artifacts are extracted.
      manual_dir: `str`, path to manually downloaded/extracted data directory.
      dataset_name: `str`, name of dataset this instance will be used for. If
        provided, downloads will contain which datasets they were used for.
      force_download: `bool`, default to False. If True, always [re]download.
      force_extraction: `bool`, default to False. If True, always [re]extract.
      register_checksums: `bool`, default to False. If True, dl checksums aren't
        checked, but stored into file.
    """
    self._dataset_name = dataset_name
    self._download_dir = os.path.expanduser(download_dir)
    self._extract_dir = os.path.expanduser(
        extract_dir or os.path.join(download_dir, 'extracted'))
    self._manual_dir = manual_dir and os.path.expanduser(manual_dir)
    tf.io.gfile.makedirs(self._download_dir)
    tf.io.gfile.makedirs(self._extract_dir)
    self._force_download = force_download
    self._force_extraction = force_extraction
    self._extractor = extractor.get_extractor()
    self._downloader = downloader.get_downloader()
    self._register_checksums = register_checksums
    # All known URLs: {url: (size, checksum)}
    self._sizes_checksums = checksums.get_all_sizes_checksums()
    # To record what is being used: {url: (size, checksum)}
    self._recorded_sizes_checksums = {}

  @property
  def downloaded_size(self):
    """Returns the total size of downloaded files."""
    return sum(size for size, sha256 in self._recorded_sizes_checksums.values())

  def _get_final_dl_path(self, url, sha256):
    return os.path.join(self._download_dir,
                        resource_lib.get_dl_fname(url, sha256))

  @util.build_synchronize_decorator()
  def _record_sizes_checksums(self):
    """Store in file when recorded size/checksum of downloaded files."""
    checksums.store_checksums(self._dataset_name,
                              self._recorded_sizes_checksums)

  def _handle_download_result(self, resource, tmp_dir_path, sha256, dl_size):
    """Store dled file to definitive place, write INFO file, return path."""
    fnames = tf.io.gfile.listdir(tmp_dir_path)
    if len(fnames) > 1:
      raise AssertionError('More than one file in %s.' % tmp_dir_path)
    original_fname = fnames[0]
    tmp_path = os.path.join(tmp_dir_path, original_fname)
    self._recorded_sizes_checksums[resource.url] = (dl_size, sha256)
    if self._register_checksums:
      self._record_sizes_checksums()
    elif (dl_size, sha256) != self._sizes_checksums.get(resource.url, None):
      raise NonMatchingChecksumError(resource.url, tmp_path)
    download_path = self._get_final_dl_path(resource.url, sha256)
    resource_lib.write_info_file(resource, download_path, self._dataset_name,
                                 original_fname)
    # Unconditionally overwrite because either file doesn't exist or
    # FORCE_DOWNLOAD=true
    tf.io.gfile.rename(tmp_path, download_path, overwrite=True)
    tf.io.gfile.rmtree(tmp_dir_path)
    return download_path

  # synchronize and memoize decorators ensure same resource will only be
  # processed once, even if passed twice to download_manager.
  @util.build_synchronize_decorator()
  @utils.memoize()
  def _download(self, resource):
    """Download resource, returns Promise->path to downloaded file."""
    if isinstance(resource, six.string_types):
      resource = resource_lib.Resource(url=resource)
    url = resource.url
    if url in self._sizes_checksums:
      expected_sha256 = self._sizes_checksums[url][1]
      download_path = self._get_final_dl_path(url, expected_sha256)
      if not self._force_download and resource.exists_locally(download_path):
        logging.info('URL %s already downloaded: reusing %s.',
                     url, download_path)
        self._recorded_sizes_checksums[url] = self._sizes_checksums[url]
        return promise.Promise.resolve(download_path)
    # There is a slight difference between downloader and extractor here:
    # the extractor manages its own temp directory, while the DownloadManager
    # manages the temp directory of downloader.
    download_dir_path = os.path.join(
        self._download_dir,
        '%s.tmp.%s' % (resource_lib.get_dl_dirname(url), uuid.uuid4().hex))
    tf.io.gfile.makedirs(download_dir_path)
    logging.info('Downloading %s into %s...', url, download_dir_path)

    def callback(val):
      checksum, dl_size = val
      return self._handle_download_result(
          resource, download_dir_path, checksum, dl_size)
    return self._downloader.download(url, download_dir_path).then(callback)

  @util.build_synchronize_decorator()
  @utils.memoize()
  def _extract(self, resource):
    """Extract a single archive, returns Promise->path to extraction result."""
    if isinstance(resource, six.string_types):
      resource = resource_lib.Resource(path=resource)
    path = resource.path
    extract_method = resource.extract_method
    if extract_method == resource_lib.ExtractMethod.NO_EXTRACT:
      logging.info('Skipping extraction for %s (method=NO_EXTRACT).', path)
      return promise.Promise.resolve(path)
    method_name = resource_lib.ExtractMethod(extract_method).name
    extract_path = os.path.join(self._extract_dir,
                                '%s.%s' % (method_name, os.path.basename(path)))
    if not self._force_extraction and tf.io.gfile.exists(extract_path):
      logging.info('Reusing extraction of %s at %s.', path, extract_path)
      return promise.Promise.resolve(extract_path)
    return self._extractor.extract(path, extract_method, extract_path)

  @util.build_synchronize_decorator()
  @utils.memoize()
  def _download_extract(self, resource):
    """Download-extract `Resource` or url, returns Promise->path."""
    if isinstance(resource, six.string_types):
      resource = resource_lib.Resource(url=resource)
    def callback(path):
      resource.path = path
      return self._extract(resource)
    return self._download(resource).then(callback)

  def download_kaggle_data(self, competition_name):
    """Download data for a given Kaggle competition."""
    with self._downloader.tqdm():
      kaggle_downloader = self._downloader.kaggle_downloader(competition_name)
      urls = kaggle_downloader.competition_urls
      files = kaggle_downloader.competition_files
      return _map_promise(self._download,
                          dict((f, u) for (f, u) in zip(files, urls)))

  def download(self, url_or_urls):
    """Download given url(s).

    Args:
      url_or_urls: url or `list`/`dict` of urls to download and extract. Each
        url can be a `str` or `tfds.download.Resource`.

    Returns:
      downloaded_path(s): `str`, The downloaded paths matching the given input
        url_or_urls.
    """
    # Add progress bar to follow the download state
    with self._downloader.tqdm():
      return _map_promise(self._download, url_or_urls)

  def iter_archive(self, resource):
    """Returns iterator over files within archive.

    **Important Note**: caller should read files as they are yielded.
    Reading out of order is slow.

    Args:
      resource: path to archive or `tfds.download.Resource`.

    Returns:
      Generator yielding tuple (path_within_archive, file_obj).
    """
    if isinstance(resource, six.string_types):
      resource = resource_lib.Resource(path=resource)
    return extractor.iter_archive(resource.path, resource.extract_method)

  def extract(self, path_or_paths):
    """Extract given path(s).

    Args:
      path_or_paths: path or `list`/`dict` of path of file to extract. Each
        path can be a `str` or `tfds.download.Resource`.

    If not explicitly specified in `Resource`, the extraction method is deduced
    from downloaded file name.

    Returns:
      extracted_path(s): `str`, The extracted paths matching the given input
        path_or_paths.
    """
    # Add progress bar to follow the download state
    with self._extractor.tqdm():
      return _map_promise(self._extract, path_or_paths)

  def download_and_extract(self, url_or_urls):
    """Download and extract given url_or_urls.

    Is roughly equivalent to:

    ```
    extracted_paths = dl_manager.extract(dl_manager.download(url_or_urls))
    ```

    Args:
      url_or_urls: url or `list`/`dict` of urls to download and extract. Each
        url can be a `str` or `tfds.download.Resource`.

    If not explicitly specified in `Resource`, the extraction method will
    automatically be deduced from downloaded file name.

    Returns:
      extracted_path(s): `str`, extracted paths of given URL(s).
    """
    # Add progress bar to follow the download state
    with self._downloader.tqdm():
      with self._extractor.tqdm():
        return _map_promise(self._download_extract, url_or_urls)

  @property
  def manual_dir(self):
    """Returns the directory containing the manually extracted data."""
    if not tf.io.gfile.exists(self._manual_dir):
      raise AssertionError(
          'Manual directory {} does not exist. Create it and download/extract '
          'dataset artifacts in there.'.format(self._manual_dir))
    return self._manual_dir


# ============================================================================
# In Python 2.X, threading.Condition.wait() cannot be interrupted by SIGINT,
# unless it's given a timeout. Here we artificially give a long timeout to
# allow ctrl+C.
# This code should be deleted once python2 is no longer supported.
if sys.version_info[0] > 2:

  def _wait_on_promise(p):
    return p.get()

else:

  def _wait_on_promise(p):
    while True:
      result = p.get(sys.maxint)  # pylint: disable=g-deprecated-member-used
      if p.is_fulfilled:
        return result

# ============================================================================


def _map_promise(map_fn, all_inputs):
  """Map the function into each element and resolve the promise."""
  all_promises = utils.map_nested(map_fn, all_inputs)  # Apply the function
  res = utils.map_nested(_wait_on_promise, all_promises)
  return res



# coding=utf-8
# Copyright 2019 The TensorFlow Datasets Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

from tensorflow_datasets import testing
from tensorflow_datasets.scripts.document_datasets import schema_org

DummyMnist = testing.DummyMnist


class DocumentDatasetsTest(testing.TestCase):

  @classmethod
  def setUpClass(cls):
    super(DocumentDatasetsTest, cls).setUpClass()
    cls._tfds_tmp_dir = testing.make_tmp_dir()
    builder = DummyMnist(data_dir=cls._tfds_tmp_dir)
    builder.download_and_prepare()

  @classmethod
  def tearDownClass(cls):
    testing.rm_tmp_dir(cls._tfds_tmp_dir)

  def setUp(self):
    self.builder = DummyMnist(data_dir=self._tfds_tmp_dir)

  @testing.run_in_graph_and_eager_modes()
  def test_schema_org(self):
    schema_str = schema_org(self.builder)
    self.assertIn("http://schema.org/Dataset", schema_str)
    self.assertIn(
        '<meta itemprop="url" '
        'content="https://www.tensorflow.org'
        '/datasets/catalog/%s" />' % self.builder.name, schema_str)


if __name__ == "__main__":
  testing.test_main()



# coding=utf-8
# Copyright 2019 The TensorFlow Datasets Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Open images datasets.

https://storage.googleapis.com/openimages/web/index.html
"""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import collections
import csv
import functools
import io
import os

from absl import logging
import numpy as np
import tensorflow as tf
import tensorflow_datasets.public_api as tfds

_DESCRIPTION = '''\
Open Images is a dataset of ~9M images that have been annotated with image-level
 labels and object bounding boxes.

The training set of V4 contains 14.6M bounding boxes for 600 object classes on
1.74M images, making it the largest existing dataset with object location
annotations. The boxes have been largely manually drawn by professional
annotators to ensure accuracy and consistency. The images are very diverse and
often contain complex scenes with several objects (8.4 per image on average).
Moreover, the dataset is annotated with image-level labels spanning thousands of
classes.
'''

_CITATION = '''\
@article{OpenImages,
  author = {Alina Kuznetsova and
            Hassan Rom and
            Neil Alldrin and
            Jasper Uijlings and
            Ivan Krasin and
            Jordi Pont-Tuset and
            Shahab Kamali and
            Stefan Popov and
            Matteo Malloci and
            Tom Duerig and
            Vittorio Ferrari},
  title = {The Open Images Dataset V4: Unified image classification,
           object detection, and visual relationship detection at scale},
  year = {2018},
  journal = {arXiv:1811.00982}
}
@article{OpenImages2,
  author = {Krasin, Ivan and
            Duerig, Tom and
            Alldrin, Neil and
            Ferrari, Vittorio
            and Abu-El-Haija, Sami and
            Kuznetsova, Alina and
            Rom, Hassan and
            Uijlings, Jasper and
            Popov, Stefan and
            Kamali, Shahab and
            Malloci, Matteo and
            Pont-Tuset, Jordi and
            Veit, Andreas and
            Belongie, Serge and
            Gomes, Victor and
            Gupta, Abhinav and
            Sun, Chen and
            Chechik, Gal and
            Cai, David and
            Feng, Zheyun and
            Narayanan, Dhyanesh and
            Murphy, Kevin},
  title = {OpenImages: A public dataset for large-scale multi-label and
           multi-class image classification.},
  journal = {Dataset available from
             https://storage.googleapis.com/openimages/web/index.html},
  year={2017}
}
'''

# Reading from .tar.gz is slower than extracting the gz and then reading from
# tar. We still read from the tar because it's faster to read fewer files on
# many network based FS.
# pylint: disable=line-too-long
_URLS = {
    'train_images': [tfds.download.Resource(  # pylint:disable=g-complex-comprehension
        url='http://open-images-dataset.s3.amazonaws.com/tar/train_%s.tar.gz' % i_,
        extract_method=tfds.download.ExtractMethod.GZIP)
                     for i_ in '0123456789abcdef'],
    'test_images': tfds.download.Resource(
        url='http://open-images-dataset.s3.amazonaws.com/tar/test.tar.gz',
        extract_method=tfds.download.ExtractMethod.GZIP),
    'validation_images': tfds.download.Resource(
        url='http://open-images-dataset.s3.amazonaws.com/tar/validation.tar.gz',
        extract_method=tfds.download.ExtractMethod.GZIP),
    'train_human_labels': 'https://storage.googleapis.com/openimages/2018_04/train/train-annotations-human-imagelabels.csv',
    'train_machine_labels': 'https://storage.googleapis.com/openimages/2018_04/train/train-annotations-machine-imagelabels.csv',
    'test_human_labels': 'https://storage.googleapis.com/openimages/2018_04/test/test-annotations-human-imagelabels.csv',
    'test_machine_labels': 'https://storage.googleapis.com/openimages/2018_04/test/test-annotations-machine-imagelabels.csv',
    'validation_human_labels': 'https://storage.googleapis.com/openimages/2018_04/validation/validation-annotations-human-imagelabels.csv',
    'validation_machine_labels': 'https://storage.googleapis.com/openimages/2018_04/validation/validation-annotations-machine-imagelabels.csv',
    'train-annotations-bbox': 'https://storage.googleapis.com/openimages/2018_04/train/train-annotations-bbox.csv',
    'test-annotations-bbox': 'https://storage.googleapis.com/openimages/2018_04/test/test-annotations-bbox.csv',
    'validation-annotations-bbox': 'https://storage.googleapis.com/openimages/2018_04/validation/validation-annotations-bbox.csv',
}
# pylint: enable=line-too-long

_Object = collections.namedtuple('Object', ['label', 'confidence', 'source'])
_Bbox = collections.namedtuple('Bbox', [
    'label', 'source', 'bbox', 'is_occluded',
    'is_truncated', 'is_group_of', 'is_depiction', 'is_inside'])

IMAGE_LEVEL_SOURCES = [
    'verification', 'crowdsource-verification',  # human labels
    'machine',
]

BBOX_SOURCES = [
    'freeform', 'xclick',  # Manually drawn boxes.
    'activemil',  # Machine generated, human controlled.
]


class OpenImagesV4Config(tfds.core.BuilderConfig):
  """BuilderConfig for OpenImagesV4."""

  def __init__(self, target_pixels=None, **kwargs):
    """BuilderConfig for OpenImagesV4.

    Args:
      target_pixels: If given, rescale the images so that the number of pixels
        is roughly this value.
      **kwargs: keyword arguments forward to super.
    """
    kwargs['supported_versions'] = [
        tfds.core.Version('2.0.0'),
        tfds.core.Version('1.0.0'),
    ]
    # Version history:
    # 2.0.0: S3 with new hashing function (different shuffle).
    # 1.0.0: S3 (new shuffling, sharding and slicing mechanism).
    super(OpenImagesV4Config, self).__init__(**kwargs)
    self._target_pixels = target_pixels

  @property
  def target_pixels(self):
    return self._target_pixels


class OpenImagesV4(tfds.core.GeneratorBasedBuilder):
  """Open Images v4."""

  BUILDER_CONFIGS = [
      OpenImagesV4Config(
          name='original',
          version=tfds.core.Version(
              '0.2.0', experiments={tfds.core.Experiment.S3: False}),
          description='Images at their original resolution and quality.'),
      OpenImagesV4Config(
          name='300k',
          version=tfds.core.Version(
              '0.2.1', experiments={tfds.core.Experiment.S3: False}),
          description='Images have roughly 300,000 pixels, at 72 JPEG quality.',
          target_pixels=300000),
      OpenImagesV4Config(
          name='200k',
          version=tfds.core.Version(
              '0.2.1', experiments={tfds.core.Experiment.S3: False}),
          description='Images have roughly 200,000 pixels, at 72 JPEG quality.',
          target_pixels=200000)
  ]

  def _info(self):
    source_class_label = tfds.features.ClassLabel(
        names=IMAGE_LEVEL_SOURCES + BBOX_SOURCES)
    all_class_label = tfds.features.ClassLabel(
        names_file=tfds.core.get_tfds_path(
            os.path.join('image', 'open_images_classes_all.txt')))
    trainable_class_label = tfds.features.ClassLabel(
        names_file=tfds.core.get_tfds_path(
            os.path.join('image', 'open_images_classes_trainable.txt')))
    boxable_class_label = tfds.features.ClassLabel(
        names_file=tfds.core.get_tfds_path(
            os.path.join('image', 'open_images_classes_boxable.txt')))
    return tfds.core.DatasetInfo(
        builder=self,
        description=_DESCRIPTION,
        features=tfds.features.FeaturesDict({
            'image': tfds.features.Image(),
            'image/filename': tfds.features.Text(),  # eg '226f0a1873b9bf8e.jpg'
            'objects': tfds.features.Sequence({
                'label': all_class_label,
                # Original data is 0, .1, ..., 1. We use 0, 1, 2, ..., 10.
                'confidence': tf.int32,
                'source': source_class_label,
            }),
            'objects_trainable': tfds.features.Sequence({
                'label': trainable_class_label,
                # Original data is 0, .1, ..., 1. We use 0, 1, 2, ..., 10.
                'confidence': tf.int32,
                'source': source_class_label,
            }),
            'bobjects': tfds.features.Sequence({
                'label': boxable_class_label,
                'source': source_class_label,
                'bbox': tfds.features.BBoxFeature(),
                # Following values can be: 1 (true), 0 (false) and -1 (unknown).
                'is_occluded': tf.int8,
                'is_truncated': tf.int8,
                'is_group_of': tf.int8,
                'is_depiction': tf.int8,
                'is_inside': tf.int8,
            }),
        }),
        urls=['https://storage.googleapis.com/openimages/web/index.html'],
        citation=_CITATION,
    )

  def _split_generators(self, dl_manager):
    """Returns SplitGenerators."""
    paths = dl_manager.download_and_extract(_URLS)
    # Load labels from CSVs:
    def load(names):
      csv_positions = [0] * len(names)
      return functools.partial(_load_objects, [paths[name] for name in names],
                               csv_positions)
    train_objects = load(['train_human_labels', 'train_machine_labels'])
    test_objects = load(['test_human_labels', 'test_machine_labels'])
    validation_objects = load(['validation_human_labels',
                               'validation_machine_labels'])
    def load_boxes(name):
      csv_positions = [0]
      return functools.partial(_load_bboxes, paths[name], csv_positions)
    train_bbox = load_boxes('train-annotations-bbox')
    test_bbox = load_boxes('test-annotations-bbox')
    validation_bbox = load_boxes('validation-annotations-bbox')
    return [
        tfds.core.SplitGenerator(
            name=tfds.Split.TRAIN,
            num_shards=512,
            gen_kwargs=dict(archive_paths=paths['train_images'],
                            objects_getter=train_objects,
                            bboxes_getter=train_bbox,
                            prefixes='0123456789abcdef'),
        ),
        tfds.core.SplitGenerator(
            name=tfds.Split.TEST,
            num_shards=36,
            gen_kwargs=dict(archive_paths=[paths['test_images']],
                            objects_getter=test_objects,
                            bboxes_getter=test_bbox),
        ),
        tfds.core.SplitGenerator(
            name=tfds.Split.VALIDATION,
            num_shards=12,
            gen_kwargs=dict(archive_paths=[paths['validation_images']],
                            objects_getter=validation_objects,
                            bboxes_getter=validation_bbox),
        ),
    ]

  def _generate_examples(self, archive_paths, objects_getter, bboxes_getter,
                         prefixes=None):
    """Yields examples."""
    trainable_classes = set(
        self.info.features['objects_trainable']['label'].names)
    for i, archive_path in enumerate(archive_paths):
      prefix = prefixes[i] if prefixes else None
      objects = objects_getter(prefix)
      bboxes = bboxes_getter(prefix)
      logging.info('Opening archive %s ...', archive_path)
      archive = tfds.download.iter_archive(
          archive_path, tfds.download.ExtractMethod.TAR_STREAM)
      for fpath, fobj in archive:
        fname = os.path.basename(fpath)
        image_id = int(os.path.splitext(fname)[0], 16)
        image_objects = [obj._asdict() for obj in objects.get(image_id, [])]
        image_bboxes = [bbox._asdict() for bbox in bboxes.get(image_id, [])]
        image_objects_trainable = [
            obj for obj in image_objects if obj['label'] in trainable_classes
        ]
        record = {
            'image': _resize_image_if_necessary(
                fobj, target_pixels=self.builder_config.target_pixels),
            'image/filename': fname,
            'objects': image_objects,
            'objects_trainable': image_objects_trainable,
            'bobjects': image_bboxes,
        }
        yield fname, record


def _resize_image_if_necessary(image_fobj, target_pixels=None):
  """Resize an image to have (roughly) the given number of target pixels.

  Args:
    image_fobj: File object containing the original image.
    target_pixels: If given, number of pixels that the image must have.

  Returns:
    A file object.
  """
  if target_pixels is None:
    return image_fobj

  cv2 = tfds.core.lazy_imports.cv2
  # Decode image using OpenCV2.
  image = cv2.imdecode(
      np.fromstring(image_fobj.read(), dtype=np.uint8), flags=3)
  # Get image height and width.
  height, width, _ = image.shape
  actual_pixels = height * width
  if actual_pixels > target_pixels:
    factor = np.sqrt(target_pixels / actual_pixels)
    image = cv2.resize(image, dsize=None, fx=factor, fy=factor)
  # Encode the image with quality=72 and store it in a BytesIO object.
  _, buff = cv2.imencode('.jpg', image, [int(cv2.IMWRITE_JPEG_QUALITY), 72])
  return io.BytesIO(buff.tostring())


def _load_objects(csv_paths, csv_positions, prefix):
  """Returns objects listed within given CSV files."""
  logging.info('Loading CSVs %s from positions %s with prefix %s',
               csv_paths, csv_positions, prefix)
  objects = collections.defaultdict(list)
  for i, labels_path in enumerate(csv_paths):
    with tf.io.gfile.GFile(labels_path) as csv_f:
      if csv_positions[i] > 0:
        csv_f.seek(csv_positions[i])
      else:
        csv_f.readline()  # Drop headers
      reader = csv.reader(csv_f)
      for image_id, source, label, confidence in reader:
        if prefix and image_id[0] != prefix:
          break
        csv_positions[i] = csv_f.tell()
        image_id = int(image_id, 16)
        current_obj = _Object(label, int(float(confidence) * 10), source)
        objects[image_id].append(current_obj)
  return dict(objects)


def _load_bboxes(csv_path, csv_positions, prefix):
  """Returns bounded boxes listed within given CSV file."""
  logging.info('Loading CSVs %s from positions %s with prefix %s',
               csv_path, csv_positions, prefix)
  boxes = collections.defaultdict(list)
  with tf.io.gfile.GFile(csv_path) as csv_f:
    if csv_positions[0] > 0:
      csv_f.seek(csv_positions[0])
    else:
      csv_f.readline()  # Drop headers
    reader = csv.reader(csv_f)
    for (image_id, source, label, confidence, xmin, xmax, ymin, ymax,
         is_occluded, is_truncated, is_group_of, is_depiction, is_inside,
        ) in reader:
      if prefix and image_id[0] != prefix:
        break
      csv_positions[0] = csv_f.tell()
      image_id = int(image_id, 16)
      del confidence  # always 1 in bounding boxes.
      current_row = _Bbox(
          label, source, tfds.features.BBox(
              float(ymin), float(xmin), float(ymax), float(xmax)),
          int(is_occluded), int(is_truncated),
          int(is_group_of), int(is_depiction), int(is_inside))
      boxes[image_id].append(current_row)
  return dict(boxes)



# coding=utf-8
# Copyright 2019 The TensorFlow Datasets Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Async download API with checksum verification. No business logic."""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import contextlib
import hashlib
import io
import os
import re
import ssl
import sys
from absl import logging
import concurrent.futures
import promise
import requests
from requests.utils import extract_zipped_paths
from six.moves import urllib
import tensorflow as tf
from tensorflow_datasets.core import units
from tensorflow_datasets.core import utils
from tensorflow_datasets.core.download import kaggle
from tensorflow_datasets.core.download import util as download_util

_DRIVE_URL = re.compile(r'^https://drive\.google\.com/')


@utils.memoize()
def get_downloader(*args, **kwargs):
  return _Downloader(*args, **kwargs)


def _get_filename(response):
  content_disposition = response.headers.get('content-disposition', None)
  if content_disposition:
    match = re.findall('filename="(.+?)"', content_disposition)
    if match:
      return match[0]
  return download_util.get_file_name(response.url)


class DownloadError(Exception):
  pass


class _Downloader(object):
  """Class providing async download API with checksum validation.

  Do not instantiate this class directly. Instead, call `get_downloader()`.
  """

  def __init__(self, max_simultaneous_downloads=50, checksumer=None):
    """Init _Downloader instance.

    Args:
      max_simultaneous_downloads: `int`, max number of simultaneous downloads.
      checksumer: `hashlib.HASH`. Defaults to `hashlib.sha256`.
    """
    self._executor = concurrent.futures.ThreadPoolExecutor(
        max_workers=max_simultaneous_downloads)
    self._checksumer = checksumer or hashlib.sha256
    self._pbar_url = None
    self._pbar_dl_size = None

  @utils.memoize()
  def kaggle_downloader(self, competition_name):
    return kaggle.KaggleCompetitionDownloader(competition_name)

  @contextlib.contextmanager
  def tqdm(self):
    """Add a progression bar for the current download."""
    async_tqdm = utils.async_tqdm
    with async_tqdm(total=0, desc='Dl Completed...', unit=' url') as pbar_url:
      with async_tqdm(total=0, desc='Dl Size...', unit=' MiB') as pbar_dl_size:
        self._pbar_url = pbar_url
        self._pbar_dl_size = pbar_dl_size
        yield

  def download(self, url, destination_path):
    """Download url to given path.

    Returns Promise -> sha256 of downloaded file.

    Args:
      url: address of resource to download.
      destination_path: `str`, path to directory where to download the resource.

    Returns:
      Promise obj -> (`str`, int): (downloaded object checksum, size in bytes).
    """
    self._pbar_url.update_total(1)
    future = self._executor.submit(self._sync_download, url, destination_path)
    return promise.Promise.resolve(future)

  def _sync_kaggle_download(self, kaggle_url, destination_path):
    """Download with Kaggle API."""
    kaggle_file = kaggle.KaggleFile.from_url(kaggle_url)
    downloader = self.kaggle_downloader(kaggle_file.competition)
    filepath = downloader.download_file(kaggle_file.filename, destination_path)

    dl_size = tf.io.gfile.stat(filepath).length
    checksum = self._checksumer()
    with tf.io.gfile.GFile(filepath, 'rb') as f:
      while True:
        block = f.read(io.DEFAULT_BUFFER_SIZE)
        if not block:
          break
        checksum.update(block)
    return checksum.hexdigest(), dl_size

  def _get_drive_url(self, url, session):
    """Returns url, possibly with confirmation token."""
    response = session.get(url, stream=True)
    if response.status_code != 200:
      raise DownloadError(
          'Failed to get url %s. HTTP code: %d.' % (url, response.status_code))
    for k, v in response.cookies.items():
      if k.startswith('download_warning'):
        return url + '&confirm=' + v  # v is the confirm token
    # No token found, let's try with original URL:
    return url

  def _sync_file_copy(self, filepath, destination_path):
    out_path = os.path.join(destination_path, os.path.basename(filepath))
    tf.io.gfile.copy(filepath, out_path)
    hexdigest, size = utils.read_checksum_digest(
        out_path, checksum_cls=self._checksumer)
    return hexdigest, size

  def _sync_download(self, url, destination_path):
    """Synchronous version of `download` method.

    Args:
      url: url to download
      destination_path: path where to write it
    Returns:
      None

    Raises:
      DownloadError: when download fails.

    Environment Variables:
        TFDS_HTTP_PROXY  : Configure Proxy Servers for HTTP Requests
        TFDS_HTTPS_PROXY : Configure Proxy Servers for HTTPS Requests
        TFDS_FTP_PROXY   : Configure Proxy Servers for FTP Requests
        TFDS_CA_BUNDLE   : Configure Custom Client Side SSL Certificates
                           (If REQUESTS_CA_BUNDLE is set, it would use that
                           path)

      Note: FTPS Custom Certificate verification doesn't work for python version
      <= 2.7.8
    """
    proxies = {
        'http': os.environ.get('TFDS_HTTP_PROXY', None),
        'https': os.environ.get('TFDS_HTTPS_PROXY', None),
        'ftp': os.environ.get('TFDS_FTP_PROXY', None)
    }
    ca_bundle = os.environ.get('TFDS_CA_BUNDLE', None)
    if not ca_bundle:
      ca_bundle = os.environ.get('REQUESTS_CA_BUNDLE', None)
    if ca_bundle:
      ca_bundle = extract_zipped_paths(ca_bundle)
    if not hasattr(ssl, '_create_unverified_context'):
      # disable ftp ssl bypassing for python version <= 2.7.8
      def disabled_py2_log_fn(*args, **kwargs):
        del args, kwargs
        return logging.info('SSL bypassing not available for python '
                            'version <= 2.7.8 current version: %s '
                            'Protocols affected: FTPS',
                            sys.version.split(' ')[0])

      ssl.__dict__['_create_unverified_context'] = disabled_py2_log_fn
      ssl.__dict__['create_default_context'] = disabled_py2_log_fn

    ca_verify = {
        'urllib':
            ssl._create_unverified_context()  # pylint: disable=W0212
            if not ca_bundle else ssl.create_default_context(capath=ca_bundle),
        'requests':
            False if not ca_bundle else ca_bundle
    }

    if kaggle.KaggleFile.is_kaggle_url(url):
      if proxies['http']:
        os.environ['KAGGLE_PROXY'] = proxies['http']
      return self._sync_kaggle_download(url, destination_path)

    try:
      # If url is on a filesystem that gfile understands, use copy. Otherwise,
      # use requests (http) or urllib (ftp).
      if not url.startswith('http'):
        return self._sync_file_copy(url, destination_path)
    except tf.errors.UnimplementedError:
      pass

    session = requests.Session()
    session.proxies = proxies
    session.verify = ca_verify['requests']
    if _DRIVE_URL.match(url):
      url = self._get_drive_url(url, session)
    use_urllib = url.startswith('ftp')
    if use_urllib:
      if proxies['ftp']:
        proxy = urllib.request.ProxyHandler({'ftp': proxies['ftp']})
        opener = urllib.request.build_opener(proxy)
        urllib.request.install_opener(opener)   # pylint: disable=too-many-function-args
      request = urllib.request.Request(url)

      # disable ssl context check for FTPS for python version <= 2.7.8
      if ca_verify['urllib'] is None:
        response = urllib.request.urlopen(request)
      else:
        response = urllib.request.urlopen(request, context=ca_verify['urllib'])
    else:
      response = session.get(url, stream=True)
      if response.status_code != 200:
        raise DownloadError('Failed to get url %s. HTTP code: %d.' %
                            (url, response.status_code))
    fname = _get_filename(response)
    path = os.path.join(destination_path, fname)
    size = 0

    size_mb = 0
    unit_mb = units.MiB
    self._pbar_dl_size.update_total(
        int(response.headers.get('Content-length', 0)) // unit_mb)
    with tf.io.gfile.GFile(path, 'wb') as file_:
      checksum = self._checksumer()
      if use_urllib:
        iterator = iter(lambda: response.read(io.DEFAULT_BUFFER_SIZE), b'')
      else:
        iterator = response.iter_content(chunk_size=io.DEFAULT_BUFFER_SIZE)

      for block in iterator:
        size += len(block)

        # Update the progress bar
        size_mb += len(block)
        if size_mb > unit_mb:
          self._pbar_dl_size.update(size_mb // unit_mb)
          size_mb %= unit_mb

        checksum.update(block)
        file_.write(block)
    self._pbar_url.update(1)
    return checksum.hexdigest(), size



# coding=utf-8
# Copyright 2019 The TensorFlow Datasets Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Bounding boxes feature."""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import collections

import tensorflow as tf

from tensorflow_datasets.core.features import feature


BBox = collections.namedtuple('BBox', 'ymin, xmin, ymax, xmax')


class BBoxFeature(feature.Tensor):
  """`FeatureConnector` for a normalized bounding box.

  Note: If you have multiple bounding boxes, you may want to wrap the feature
  inside a `tfds.feature.Sequence`.

  Input:
    * `tfds.features.BBox` tuple.

  Output:
    bbox: tf.Tensor of type tf.float32 and shape [4,] which contains the
      normalized coordinates of the bounding box [ymin, xmin, ymax, xmax]

  Example:
    * In the DatasetInfo object:
      features=features.FeatureDict({
          'bbox': features.BBox(shape=(None, 64, 64, 3)),
      })

    * During generation:
      yield {
          'input': tfds.feature.BBox(ymin=0.3, xmin=0.8, ymax=0.5, xmax=1.0),
      }
  """

  def __init__(self):
    super(BBoxFeature, self).__init__(shape=(4,), dtype=tf.float32)

  def encode_example(self, bbox):
    """See base class for details."""
    # Validate the coordinates
    for coordinate in bbox:
      if not isinstance(coordinate, float):
        raise ValueError(
            'BBox coordinates should be float. Got {}.'.format(bbox))
      if not 0.0 <= coordinate <= 1.0:
        raise ValueError(
            'BBox coordinates should be between 0 and 1. Got {}.'.format(bbox))
      if bbox.xmax < bbox.xmin or bbox.ymax < bbox.ymin:
        raise ValueError(
            'BBox coordinates should have min <= max. Got {}.'.format(bbox))

    return super(BBoxFeature, self).encode_example(
        [bbox.ymin, bbox.xmin, bbox.ymax, bbox.xmax]
    )



# coding=utf-8
# Copyright 2019 The TensorFlow Datasets Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

r"""Generate Smallnorb-like files, smaller and with random data.

"""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import os

from absl import app
from absl import flags
import numpy as np
import tensorflow as tf

from tensorflow_datasets.core.utils import py_utils
from tensorflow_datasets.testing import test_utils

NUM_IMAGES = 5
FACTOR_VALUES = [
    list(range(10)),
    list(range(9)),
    list(range(0, 36, 2)),
    list(range(6)),
]
TRAINING_OUTPUT_NAME = "smallnorb-5x46789x9x18x6x2x96x96-training"
TESTING_OUTPUT_NAME = "smallnorb-5x01235x9x18x6x2x96x96-testing"

flags.DEFINE_string("tfds_dir", py_utils.tfds_dir(),
                    "Path to tensorflow_datasets directory.")
FLAGS = flags.FLAGS


def write_binary_matrix(filename, array):
  """Writes array as a binary formatted matrix to the file.

  The file format is described on the data set page:
  https://cs.nyu.edu/~ylclab/data/norb-v1.0-small/

  Args:
    filename: String with path to the file.
    array: Numpy array that should be written to the file.
  """
  with tf.io.gfile.GFile(filename, "wb") as f:

    # All data is stored in little-endian byte order.
    int32_dtype = np.dtype("int32").newbyteorder("<")

    # The first 4 bytes specify the data type.
    if array.dtype.str == "<i4":
      # Magic code for little-endian int32.
      f.write(np.asarray(507333716, dtype=int32_dtype).tobytes())
    elif array.dtype.str == "|u1":
      # Magic code for uint8.
      f.write(np.asarray(507333717, dtype=int32_dtype).tobytes())
    else:
      raise ValueError("Array data type %r not supported." % array.dtype.str)

    # Next, we specify the number of dimensions of the array to be stored as a
    # 32 bit integer.
    f.write(np.asarray(array.ndim, int32_dtype).tobytes())

    # The shape of the array is saved as 32-bit integers. If there are less than
    # 3 dimensions, the shape is padded with ones.
    shape = list(array.shape) + [1] * max(0, 3 - array.ndim)
    f.write(np.asarray(shape, int32_dtype).tobytes())

    # Finally, the data is written as a C-contiguous matrix. There is no need to
    # check for the byte order as we checked for this when writing the magic
    # code.
    f.write(np.ascontiguousarray(array).tobytes())


def _create_chunk(prefix, random_state):
  """Creates fake dat, cat, and info files with the given prefix."""
  # Create the images.
  image_shape = (NUM_IMAGES, 2, 96, 96)  # Data file contains pairs of images.
  images = random_state.randint(256, size=image_shape).astype("uint8")
  write_binary_matrix("%s-dat.mat" % prefix, images)

  # Create the class label file.
  class_labels = random_state.choice(range(5), size=(NUM_IMAGES))
  write_binary_matrix("%s-cat.mat" % prefix, class_labels.astype("int32"))

  # Create the auxiliary info file that contains additional labels.
  info = []
  for values in FACTOR_VALUES:
    info.append(random_state.choice(values, size=(NUM_IMAGES)))
  write_binary_matrix("%s-info.mat" % prefix, np.array(info).T.astype("int32"))


def _generate():
  """Generates a fake data set and writes it to the fake_examples directory."""
  output_dir = os.path.join(FLAGS.tfds_dir, "testing", "test_data",
                            "fake_examples", "smallnorb")
  test_utils.remake_dir(output_dir)
  random_state = np.random.RandomState(0)
  _create_chunk(os.path.join(output_dir, TRAINING_OUTPUT_NAME), random_state)
  _create_chunk(os.path.join(output_dir, TESTING_OUTPUT_NAME), random_state)


def main(argv):
  if len(argv) > 1:
    raise app.UsageError("Too many command-line arguments.")
  _generate()


if __name__ == "__main__":
  app.run(main)



# coding=utf-8
# Copyright 2019 The TensorFlow Datasets Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""To serialize Dict or sequence to Example."""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import numpy as np
import six
import tensorflow as tf

from tensorflow_datasets.core import utils


class ExampleSerializer(object):
  """To serialize examples."""

  def __init__(self, example_specs):
    """Constructor.

    Args:
      example_specs: Nested `dict` of `tfds.features.TensorInfo`, corresponding
        to the structure of data to write/read.
    """
    self._example_specs = example_specs
    self._flat_example_specs = utils.flatten_nest_dict(self._example_specs)

  def serialize_example(self, example):
    """Serialize the given example.

    Args:
      example: Nested `dict` containing the input to serialize. The input
        structure and values dtype/shape must match the `example_specs`
        provided at construction.

    Returns:
      serialize_proto: `str`, the serialized `tf.train.Example` proto
    """
    example = utils.flatten_nest_dict(example)
    example = _dict_to_tf_example(example, self._flat_example_specs)
    return example.SerializeToString()


def _dict_to_tf_example(example_dict, tensor_info_dict=None):
  """Builds tf.train.Example from (string -> int/float/str list) dictionary.

  Args:
    example_dict: `dict`, dict of values, tensor,...
    tensor_info_dict: `dict` of `tfds.feature.TensorInfo` If given, perform
      additional checks on the example dict (check dtype, shape, number of
      fields...)
  """
  def serialize_single_field(k, example_data, tensor_info):
    with utils.try_reraise(
        "Error while serializing feature {} ({}): ".format(k, tensor_info)):
      return _item_to_tf_feature(example_data, tensor_info)

  if tensor_info_dict:
    example_dict = {
        k: serialize_single_field(k, example_data, tensor_info)
        for k, (example_data, tensor_info)
        in utils.zip_dict(example_dict, tensor_info_dict)
    }
  else:
    example_dict = {
        k: serialize_single_field(k, example_data, None)
        for k, example_data in example_dict.items()
    }

  return tf.train.Example(features=tf.train.Features(feature=example_dict))


def _is_string(item):
  """Check if the object contains string or bytes."""
  if isinstance(item, (six.binary_type, six.string_types)):
    return True
  elif (isinstance(item, (tuple, list)) and
        all(isinstance(x, (six.binary_type, six.string_types)) for x in item)):
    return True
  elif (isinstance(item, np.ndarray) and  # binary or unicode
        (item.dtype.kind in ("U", "S") or item.dtype == object)):
    return True
  return False


def _item_to_tf_feature(item, tensor_info=None):
  """Single item to a tf.train.Feature."""
  v = item
  if not tensor_info and isinstance(v, (list, tuple)) and not v:
    raise ValueError(
        "Received an empty list value, so is unable to infer the "
        "feature type to record. To support empty value, the corresponding "
        "FeatureConnector should return a numpy array with the correct dtype "
        "instead of a Python list."
    )

  # Handle strings/bytes first
  is_string = _is_string(v)

  if tensor_info:
    np_dtype = np.dtype(tensor_info.dtype.as_numpy_dtype)
  elif is_string:
    np_dtype = object  # Avoid truncating trailing '\x00' when converting to np
  else:
    np_dtype = None

  v = np.array(v, dtype=np_dtype)

  # Check that the shape is expected
  if tensor_info:
    utils.assert_shape_match(v.shape, tensor_info.shape)
    if tensor_info.dtype == tf.string and not is_string:
      raise ValueError(
          "Unsuported value: {}\nCould not convert to bytes list.".format(item))

  # Convert boolean to integer (tf.train.Example does not support bool)
  if v.dtype == np.bool_:
    v = v.astype(int)

  v = v.flatten()  # Convert v into a 1-d array
  if np.issubdtype(v.dtype, np.integer):
    return tf.train.Feature(int64_list=tf.train.Int64List(value=v))
  elif np.issubdtype(v.dtype, np.floating):
    return tf.train.Feature(float_list=tf.train.FloatList(value=v))
  elif is_string:
    v = [tf.compat.as_bytes(x) for x in v]
    return tf.train.Feature(bytes_list=tf.train.BytesList(value=v))
  else:
    raise ValueError(
        "Unsuported value: {}.\n"
        "tf.train.Feature does not support type {}. "
        "This may indicate that one of the FeatureConnectors received an "
        "unsupported value as input.".format(repr(v), repr(type(v)))
    )
